# -*- coding: utf-8 -*-
"""
http://pythonhosted.org/psycopg2/module.html

http://zetcode.com/db/postgresqlpythontutorial/
http://wiki.postgresql.org/wiki/Using_psycopg2_with_PostgreSQL
http://pythonhosted.org/psycopg2/
"""
from __future__ import unicode_literals, division, print_function, absolute_import
import re
import os
import sys
import decimal
import datetime
import uuid
import json
from collections import Counter

# third party
import psycopg2
import psycopg2.extras
import psycopg2.extensions

# first party
from ..sql import SQLInterface, SQLConnection
from ...compat import *
from ...utils import get_objects
from ...exception import (
    InterfaceError,
    UniqueError,
    TableError,
    FieldError,
    UniqueError,
    CloseError,
    PlaceholderError,
)


# class LoggingCursor(psycopg2.extras.RealDictCursor):
#     def execute(self, sql, args=None):
#         logger.debug(self.mogrify(sql, args))
#         super(LoggingCursor, self).execute(sql, args)
#         #psycopg2.extensions.cursor.execute(self, sql, args)


class DictType(psycopg2.extensions.QuotedString):
    """Converts from python dict to JSONB to be saved into the db

    https://www.psycopg.org/docs/extensions.html#psycopg2.extensions.QuotedString
    https://www.psycopg.org/docs/extensions.html#psycopg2.extensions.register_adapter

    Surprisingly, Postgres converts the json value in the db back to dict just fine
    but does not adapt a dict, I'm not sure why
    """
    @classmethod
    def adapt(cls, val):
        """adapter should be a function taking a single argument (the object to adapt)
        and returning an object conforming to the ISQLQuote protocol (e.g. exposing
        a getquoted() method). Once an object is registered, it can be safely used
        in SQL queries and by the adapt() function

        :param val: dict, the value coming from Python destined for Postgres
        :returns: str
        """
        return cls(json.dumps(val))


class StringType(object):
    @classmethod
    def convert(cls, val, cur):
        """Convert a db string to a python str type

        https://www.psycopg.org/docs/extensions.html#psycopg2.extensions.new_type

            These functions are used to manipulate type casters to convert from
            PostgreSQL types to Python objects.

            adapter should have signature fun(value, cur) where value is the string
            representation returned by PostgreSQL and cur is the cursor from which
            data are read. In case of NULL, value will be None. The adapter should
            return the converted object.

        :param val: str, the value coming from Postgres and destined for Python
        :param cur: cursor
        :returns: str
        """
        if isinstance(val, str) and val.startswith("\\x"):
            buf = psycopg2.BINARY(val, cur)
            val = bytes(buf).decode(cur.connection.encoding)

        return val


class PostgreSQLConnection(SQLConnection, psycopg2.extensions.connection):
#class Connection(SQLConnection, psycopg2.extras.LoggingConnection):
    """
    https://www.psycopg.org/docs/connection.html
    http://initd.org/psycopg/docs/advanced.html
    http://initd.org/psycopg/docs/extensions.html#psycopg2.extensions.connection
    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # http://initd.org/psycopg/docs/connection.html#connection.autocommit
        self.autocommit = True

        psycopg2.extensions.register_type(
            # https://www.psycopg.org/docs/extensions.html#psycopg2.extensions.new_type
            psycopg2.extensions.new_type(psycopg2.STRING.values, "STRING", StringType.convert)
        )

        # https://www.psycopg.org/docs/extensions.html#psycopg2.extensions.register_adapter
        psycopg2.extensions.register_adapter(dict, DictType.adapt)


class PostgreSQL(SQLInterface):
    """
    https://www.psycopg.org/docs/
    https://www.psycopg.org/docs/usage.html
    """
    LIMIT_NONE = "ALL"

    _connection_pool = None

    @classmethod
    def configure(cls, config):
        config = super().configure(config)
        port = config.port
        if not port:
            config.port = 5432
        return config

    def get_paramstyle(self):
        """
        https://www.psycopg.org/docs/module.html#psycopg2.paramstyle
        """
        return psycopg2.paramstyle

    def _connect(self, config):
        # set pool_key to None to use the pool as an actual pool, right now we've
        # turned it back into basically a sync interface since I don't need the gevent
        # stuff right now and I want to switch to using the native async interface
        self.config.options.setdefault("pool_key", "sync_interface")
        minconn = int(config.options.get('pool_minconn', 1))
        maxconn = int(config.options.get('pool_maxconn', 1))
        pool_class_name = config.options.get(
            'pool_class',
            'psycopg2.pool.SimpleConnectionPool'
        )

        _, pool_class = get_objects(pool_class_name)

        self.log_debug("connecting using pool class {}".format(pool_class_name))

        # https://www.psycopg.org/docs/pool.html
        # http://initd.org/psycopg/docs/module.html#psycopg2.connect
        self._connection_pool = pool_class(
            minconn,
            maxconn,
            dbname=config.database,
            user=config.username,
            password=config.password,
            host=config.host,
            port=config.port,
            cursor_factory=psycopg2.extras.RealDictCursor,
            #cursor_factory=LoggingCursor,
            connection_factory=PostgreSQLConnection,
        )

        self._connection_pool.keys = Counter()

    def _get_connection(self):
        key = self.config.options["pool_key"]
        connection = self._connection_pool.getconn(
            key=key,
        )

        connection_id = id(connection)

        # change the connection readonly status if they don't match
        if bool(connection.readonly) != bool(self.config.readonly):
            # https://www.psycopg.org/docs/connection.html#connection.readonly
            self.log_warning([
                f"Changing connection {connection_id:02x}",
                f"to readonly={self.config.readonly}",
                f"from readonly={connection.readonly}",
            ])
            connection.readonly = self.config.readonly

        self._connection_pool.keys[key] += 1

        return connection

    def _free_connection(self, connection):
        # if an error was handled there is a chance that the connection was reset
        # and we don't want to put a dead connection back into the pool
        key = self.config.options["pool_key"]
        self._connection_pool.keys[key] -= 1

        if connection.closed:
            try:
                self.log_warning(f"Discarding pool connection {id(connection):02x} because it is closed")
                self._connection_pool.putconn(
                    connection,
                    key=key,
                    close=True,
                )

            except KeyError:
                # something happened and so we just discard this instead of
                # trying to clean it up
                pass

        else:
            if self._connection_pool.keys[key] <= 0:
                self._connection_pool.putconn(
                    connection,
                    key=key,
                )

    def _close(self):
        self._connection_pool.closeall()
        self._connection_pool = None

    def _readonly(self, readonly, **kwargs):
        """readonly setting is handled when you grab the connection from get_connection()
        so this method does nothing"""
        pass

    def _get_tables(self, table_name, **kwargs):
        query_str = 'SELECT tablename FROM pg_tables WHERE tableowner = %s'
        query_args = [self.config.username]

        if table_name:
            query_str += ' AND tablename = %s'
            query_args.append(str(table_name))

        schema_name = kwargs.get("schema_name", "public")
        if schema_name:
            query_str += ' AND schemaname = %s'
            query_args.append(schema_name)

        ret = self._raw(query_str, *query_args, **kwargs)
        # http://www.postgresql.org/message-id/CA+mi_8Y6UXtAmYKKBZAHBoY7F6giuT5WfE0wi3hR44XXYDsXzg@mail.gmail.com
        return [r['tablename'] for r in ret]

    def _inserts(self, schema, field_names, field_values, **kwargs):
        """
        https://www.psycopg.org/docs/cursor.html#cursor.executemany
        https://www.psycopg.org/docs/extras.html#fast-exec
        https://www.psycopg.org/docs/extras.html#psycopg2.extras.execute_batch
        """
        query_str = self.render_inserts_sql(schema, field_names, **kwargs)
        with self.connection(**kwargs) as connection:
            cur = connection.cursor()
            psycopg2.extras.execute_batch(
                cur,
                query_str,
                field_values,
            )

    def _delete_table(self, schema, **kwargs):
        """
        https://www.postgresql.org/docs/current/sql-droptable.html
        """
        query_str = 'DROP TABLE IF EXISTS {} CASCADE'.format(self.render_table_name_sql(schema))
        self._raw(query_str, ignore_result=True, **kwargs)

    def _get_fields(self, table_name, **kwargs):
        """return all the fields for the given schema"""
        ret = {}
        query_args = ['f', self.render_table_name_sql(table_name)]

        # I had to brush up on my join knowledge while writing this query
        # https://en.wikipedia.org/wiki/Join_(SQL)
        #
        # other helpful links
        # * https://wiki.postgresql.org/wiki/Retrieve_primary_key_columns
        # * https://www.postgresql.org/docs/9.4/static/catalog-pg-attribute.html
        # * https://www.postgresql.org/docs/9.3/static/catalog-pg-type.html
        # 
        # another approach
        # * http://dba.stackexchange.com/questions/22362/
        # * http://gis.stackexchange.com/questions/94049/
        query_str = [
            'SELECT',
            ',  '.join([
                'a.attnum',
                'a.attname',
                'a.attnotnull',
                't.typname',
                'i.indisprimary',
                #'s.conname',
                #'pg_get_constraintdef(s.oid, true) as condef',
                'c.relname AS confrelname',
            ]),
            'FROM',
            '  pg_attribute a',
            'JOIN pg_type t ON a.atttypid = t.oid',
            'LEFT JOIN pg_index i ON a.attrelid = i.indrelid',
            '  AND a.attnum = any(i.indkey)',
            'LEFT JOIN pg_constraint s ON a.attrelid = s.conrelid',
            '  AND s.contype = {} AND a.attnum = any(s.conkey)'.format(self.PLACEHOLDER),
            'LEFT JOIN pg_class c ON s.confrelid = c.oid',
            'WHERE',
            '  a.attrelid = {}::regclass'.format(self.PLACEHOLDER),
            '  AND a.attisdropped = False',
            '  AND a.attnum > 0',
            'ORDER BY a.attnum ASC',
        ]
        query_str = "\n".join(query_str)
        fields = self._raw(query_str, *query_args, **kwargs)

        pg_types = {
            "float4": float,
            "float8": float,
            "timestamp": datetime.datetime,
            "timestamptz": datetime.datetime,
            "int2": int,
            "int4": int,
            "int8": long,
            "numeric": decimal.Decimal,
            "text": str,
            "citext": str,
            "bpchar": str,
            "varchar": str,
            "bool": bool,
            "date": datetime.date,
            "blob": bytes,
            "bytea": bytes,
            "jsonb": bytes,
            "uuid": uuid.UUID,
        }

        # the rows we can set: field_type, name, field_required, min_size, max_size,
        #   size, unique, pk, <foreign key info>
        # These keys will roughly correspond with schema.Field
        for row in fields:
            field = {
                "name": row["attname"],
                "field_type": pg_types[row["typname"]],
                "field_required": row["attnotnull"],
                "pk": bool(row["indisprimary"]),
            }

            if row["confrelname"]:
                # TODO -- I can't decide which name I like
                field["schema_table_name"] = row["confrelname"]
                field["ref_table_name"] = row["confrelname"]

            if row["typname"] == "citext":
                field["ignore_case"] = True

            ret[field["name"]] = field

        return ret

    def _get_indexes(self, schema, **kwargs):
        """return all the indexes for the given schema"""
        ret = {}
        query_str = [
            'SELECT',
            '  tbl.relname AS table_name, i.relname AS index_name, a.attname AS field_name,',
            '  ix.indkey AS index_order, a.attnum AS field_num',
            'FROM',
            '  pg_class tbl, pg_class i, pg_index ix, pg_attribute a',
            'WHERE',
            '  tbl.oid = ix.indrelid AND i.oid = ix.indexrelid AND a.attrelid = tbl.oid',
            '  AND a.attnum = ANY(ix.indkey) AND tbl.relkind = %s AND tbl.relname = %s',
            'ORDER BY',
            '  tbl.relname, i.relname',
        ]
        query_str = "\n".join(query_str)

        indexes = self._raw(query_str, 'r', str(schema), **kwargs)

        # massage the data into more readable {index_name: fields} format
        for idict in indexes:
            if idict['index_name'] not in ret:
                ret[idict['index_name']] = list(map(int, idict['index_order'].split(' ')))

            i = ret[idict['index_name']].index(idict['field_num'])
            ret[idict['index_name']][i] = idict['field_name']

        return ret

    def render_sort_field_sql(self, field_name, field_vals, sort_dir_str):
        # this solution is based off:
        # http://postgresql.1045698.n5.nabble.com/ORDER-BY-FIELD-feature-td1901324.html
        # see also: https://gist.github.com/cpjolicoeur/3590737
        query_sort_str = []
        query_args = []
        for v in reversed(field_vals):
            query_sort_str.append('  {} = {} {}'.format(
                self.render_field_name_sql(field_name), self.PLACEHOLDER, sort_dir_str))
            query_args.append(v)

        return ',\n'.join(query_sort_str), query_args

    def render_date_field_sql(self, field_name, field_kwargs, symbol):
        """
        allow extracting information from date

        http://www.postgresql.org/docs/8.3/static/functions-datetime.html#FUNCTIONS-DATETIME-EXTRACT
        """
        fstrs = []
        k_opts = {
            'century': 'EXTRACT(CENTURY FROM {})',
            'day': 'EXTRACT(DAY FROM {})',
            'decade': 'EXTRACT(DECADE FROM {})',
            'weekday': 'EXTRACT(DOW FROM {})',
            'dow': 'EXTRACT(DOW FROM {})',
            'isodow': 'EXTRACT(ISODOW FROM {})',
            'epoch': 'EXTRACT(EPOCH FROM {})',
            'hour': 'EXTRACT(HOUR FROM {})',
            'year': 'EXTRACT(YEAR FROM {})',
            'isoyear': 'EXTRACT(ISOYEAR FROM {})',
            'minute': 'EXTRACT(MINUTE FROM {})',
            'month': 'EXTRACT(MONTH FROM {})',
            'quarter': 'EXTRACT(QUARTER FROM {})',
            'week': 'EXTRACT(WEEK FROM {})',
        }

        for k, v in field_kwargs.items():
            fstrs.append([k_opts[k].format(self.render_field_name_sql(field_name)), self.PLACEHOLDER, v])

        return fstrs

    def render_datatype_int_sql(self, field_name, field, **kwargs):
        if field.is_ref():
            field_type = 'BIGINT' # INT8

        else:
            # https://www.postgresql.org/docs/current/datatype-numeric.html
            size_info = field.size_info()
            size = size_info["size"]

            if size == 0:
                field_type = 'INTEGER' # INT4

            elif size <= 32767:
                field_type = 'SMALLINT' # INT2

            elif size <= 2147483647:
                # INT4
                field_type = 'SERIAL' if field.is_auto() else 'INTEGER'

            elif size <= 9223372036854775807:
                # INT8
                field_type = 'BIGSERIAL' if field.is_auto() else 'BIGINT'

            else:
                precision = size_info["precision"]
                field_type = f'NUMERIC({precision}, 0)'

        return field_type

    def render_datatype_str_sql(self, field_name, field, **kwargs):
        if field.interface_options.get('ignore_case', False):
            kwargs.setdefault("datatype", "CITEXT")

        return super().render_datatype_str_sql(field_name, field, **kwargs)

    def render_datatype_datetime_sql(self, field_name, field, **kwargs):
        # http://www.postgresql.org/docs/9.0/interactive/datatype-datetime.html
        #field_type = 'TIMESTAMP WITHOUT TIME ZONE'

        # https://wiki.postgresql.org/wiki/Don't_Do_This#Don.27t_use_timestamp_.28without_time_zone.29
        return 'TIMESTAMPTZ'

    def render_datatype_dict_sql(self, field_name, field, **kwargs):
        # https://www.postgresql.org/docs/current/datatype-json.html
        # In general, most applications should prefer to store JSON data as
        # jsonb, unless there are quite specialized needs
        return 'JSONB'

    def render_datatype_float_sql(self, field_name, field, **kwargs):
        """
        https://www.postgresql.org/docs/current/datatype-numeric.html#DATATYPE-NUMERIC-DECIMAL
        """
        size_info = field.size_info()
        if size_info["has_precision"]:
            precision = size_info["precision"]
            scale = size_info["scale"]
            field_type = f'NUMERIC({precision}, {scale})'

        else:
            size = size_info["size"]

            # https://learn.microsoft.com/en-us/cpp/c-language/type-float
            if size < 3.402823466e+38:
                #field_type = 'REAL'
                # REAL only has 6 digits after the period, double precision has
                # 15 digits after the period
                field_type = 'DOUBLE PRECISION'

            elif size < 1.7976931348623158e+308:
                field_type = 'DOUBLE PRECISION'

            else:
                precision = size_info["precision"]
                field_type = f'NUMERIC({precision})'

        return field_type

    def render_datatype_bytes_sql(self, field_name, field, **kwargs):
        """Why Postgres? Why?

        https://www.postgresql.org/docs/current/datatype-binary.html
            The SQL standard defines a different binary string type, called BLOB
            or BINARY LARGE OBJECT. The input format is different from bytea, but
            the provided functions and operators are mostly the same.
        """
        return 'BYTEA'

    def render_datatype_uuid_sql(self, field_name, field, **kwargs):
        # https://www.postgresql.org/docs/current/datatype-uuid.html
        # https://www.postgresql.org/docs/current/functions-uuid.html
        field_type = 'UUID'
        if field.is_auto():
            field_type += ' DEFAULT gen_random_uuid()'

        return field_type

    def create_error(self, e, **kwargs):
        """
        https://www.psycopg.org/docs/module.html#exceptions
        """
        kwargs.setdefault("error_module", psycopg2)
        if isinstance(e, psycopg2.ProgrammingError):
            e_msg = String(e)
            if "does not exist" in e_msg:
                if "column" in e_msg:
                    #INSERT: column "foo" of relation "<TABLE-NAME>" does not exist
                    #SELECT: column "foo" does not exist
                    e = FieldError(e)

                else:
                    #'relation "<TABLE-NAME>" does not exit'
                    e = TableError(e)

            elif "can't adapt type" in e_msg and "error_args" in kwargs:
                error_args = kwargs["error_args"]
                ms = re.search(r"type\s'(\S+)'", e_msg)
                type_name = ms.group(1)
                for index, value in enumerate(error_args[1]):
                    if type_name in str(type(value)):
                        e = PlaceholderError(
                            e,
                            *error_args,
                            message=f"Placeholder {index} of query has unexpected type {type(value)}",
                        )
                        break

            else:
                e = super().create_error(e, **kwargs)

        elif isinstance(e, psycopg2.errors.AdminShutdown):
            e = CloseError(e)

        elif isinstance(e, psycopg2.errors.InFailedSqlTransaction):
            e = CloseError(e)

        elif isinstance(e, psycopg2.IntegrityError):
            e = UniqueError(e)

        elif isinstance(e, IndexError) and "error_args" in kwargs:
            e = PlaceholderError(e, *kwargs.get("error_args", []))

        else:
            e = super().create_error(e, **kwargs)

        return e




# -*- coding: utf-8 -*-
"""
Classes and stuff that handle querying the interface for a passed in Orm class
"""
from __future__ import unicode_literals, division, print_function, absolute_import
import copy
from collections import defaultdict, OrderedDict
import datetime
import logging
import os
from contextlib import contextmanager
import math
import inspect
import time
import re

from datatypes.collections import ListIterator
from datatypes import property as cachedproperty

from .utils import make_list, get_objects, make_dict, make_hash
from .interface import get_interfaces
from .compat import *


logger = logging.getLogger(__name__)


class Iterator(ListIterator):
    """The main iterator for all query methods that return iterators

    This is returned from the Query.get() and Query.all() methods, this is also
    the Iterator class that is set in Orm.iterator_class

    fields --
        ifilter -- callback -- an iterator filter, all yielded rows will be passed
            through this callback and skipped if ifilter(row) returns False

    examples --
        # iterate through all the primary keys of some orm
        for pk in SomeOrm.query.all().pk:
            print pk
    """
    @property
    def orm_class(self):
        return self.query.orm_class

    def __init__(self, query):
        """create an iterator for a query

        :param query: Query, the query instance that produced this iterator
        """
        self.query = query

        if query._ifilter:
            self.ifilter = query._ifilter # https://docs.python.org/2/library/itertools.html#itertools.ifilter

        self.reset()

    def has_more(self):
        """Return true if there are more results for this query if the query didn't
        have a LIMIT clause

        :returns: boolean, True if this query could've returned more results
        """
        ret = False
        if self.query.bounds.has_more():
            cursor = self.cursor()
            # https://www.psycopg.org/docs/cursor.html#cursor.rowcount says that
            # future versions of the spec reserve the right to return None
            if cursor.rowcount == -1 or cursor.rowcount is None:
                try:
                    if self[self.query.bounds.find_more_index()]:
                        ret = True

                except IndexError:
                    pass

            else:
                ret = self.query.bounds.limit_paginate == cursor.rowcount

        return ret

    def cursor(self):
        cursor = getattr(self, "_cursor", None)
        if not cursor:
            cursor = self.query.cursor()
            self._cursor = cursor
            self._cursor_i = 0
            self.field_names = self.query.fields_select.names()

        return cursor

    def reset(self):
        """put all the pieces together to build a generator of the results"""
        self._cursor = None
        self._cursor_i = 0

    def __iter__(self):
        self.reset()
        return self

    def next(self):
        cursor = self.cursor()
        cursor_next = cursor.__next__

        # if we have paginated the results we have to account for requesting the
        # one extra row to see if we have more results waiting
        if self.query.bounds.has_more():
            #if cursor.rownumber == self.query.bounds.limit:
            if self._cursor_i == self.query.bounds.limit:
                raise StopIteration()

        o = self.hydrate(cursor_next())
        self._cursor_i += 1
        while not self.ifilter(o):
            o = self.hydrate(cursor_next())
            self._cursor_i += 1
        return o

    def count(self):
        cursor = self.cursor()
        count = cursor.rowcount

        if count >= 0:
            # compensate for having pulled one extra row
            if self.query.bounds.has_more() and self.query.bounds.limit_paginate == count:
                count -= 1

        else:
            # we couldn't get the rowcount from the cursor for some reason, so
            # we will need to query for it

            # SQLite cursor's will always have rowcount=-1 until the cursor is
            # exhausted or a fetch*() method has been called, ugh.
            # https://stackoverflow.com/a/839419/5006
            count = self.query.copy().count()

        return count

    def __getitem__(self, i):
        it = self.copy()
        q = it.query

        b = q.bounds
        limit = b.limit if b.has_limit() else self.count()
        b = q.bounds_class(limit=limit, offset=q.bounds.offset)

        if isinstance(i, slice):
            if i.step:
                raise ValueError("slice stepping is not supported")

            if i.start:
                start = b.find_offset(i.start)
            else:
                start = b.offset

            if i.stop:
                stop = b.find_offset(i.stop)
            else:
                stop = b.limit

            q.limit(stop - start).offset(start)
            it.query = q
            return it

        else:
            offset = b.find_offset(i)

            o = q.offset(offset).one()
            if o is None:
                raise IndexError("Iterator index {} out of range".format(i))

            return o

    def copy(self):
        q = self.query.copy()
        return type(self)(q)

    def reverse(self):
        for f in self.query.fields_sort:
            f.direction = -f.direction
        self.reset()

    def __getattr__(self, field_name):
        """If you have a set of results and just want to grab a certain field then
        you can do that

        :example:
            it = FooOrm.query.limit(10).get()
            it.pk # get all the primary keys from the results

        :param field_name: string, the field name you want the values of
        :returns: generator, the field_name values
        """
        it = self.copy()
        it.query.fields_select.clear()
        it.query.select_field(field_name)
        return it
        #return (getattr(o, k) for o in self)

    def __repr__(self):
        format_str = "[ ... {} ... ]"
        format_args = [self.__class__.__name__]

        orm_class = self.orm_class
        if orm_class:
            format_str = "[ ... {} on {} ... ]"
            format_args.append(orm_class.__name__)

        return format_str.format(*format_args)

    def ifilter(self, o):
        """run o through the filter, if True then orm o should be included

        NOTE -- The ifilter callback needs to account for non Orm instance values of o

        :param o: Orm|mixed, usually an Orm instance but can also be a tuple or single value
        :returns: boolean, True if o should be filtered
        """
        return o

    def hydrate(self, d):
        """Prepare the raw dict d returned from the interface cursor to be returned
        by higher level objects, this will usually mean hydrating an Orm instance
        or stuff like that

        :param d: dict, the raw dict cursor result returned from the interface
        :returns: mixed, usually an Orm instance populated with d but can also be
            a tuple if the query selected more than one field. If the query selected
            one field then just that value will be returned
        """
        r = None
        orm_class = self.orm_class
        if self.field_names:
            field_vals = []
            for field_name in self.field_names:
                fv = d[field_name]
                if orm_class:
                    fv = orm_class.schema.fields[field_name].iget(None, fv)
                field_vals.append(fv)
            r = field_vals if len(self.field_names) > 1 else field_vals[0]

        else:
            if orm_class:
                r = orm_class.hydrate(d)
            else:
                r = d

        return r


class QueryBounds(object):
    @property
    def limit(self):
        l = self._limit
        #l = self.limit_paginate if self.paginate else self._limit
        return l if l else 0

    @limit.setter
    def limit(self, v):
        v = int(v)
        if v < 0:
            raise ValueError("Limit cannot be negative")
        self._limit = v

    @limit.deleter
    def limit(self):
        self._limit = None

    @property
    def limit_paginate(self):
        limit = 0 if self._limit is None else self._limit
        return limit + 1 if limit > 0 else 0

    @property
    def offset(self):
        offset = self._offset
        if offset is None:
            page = self.page
            #limit = self.limit
            limit = self._limit
            if not limit: limit = 1
            offset = (page - 1) * limit

        else:
            offset = offset if offset >= 0 else 0

        return offset

    @offset.setter
    def offset(self, v):
        v = int(v)
        if v < 0:
            raise ValueError("Offset cannot be negative")
        del self.page
        self._offset = v

    @offset.deleter
    def offset(self):
        self._offset = None

    @property
    def page(self):
        page = 0 if self._page is None else self._page
        return page if page >= 1 else 1

    @page.setter
    def page(self, v):
        v = int(v)
        if v < 0:
            raise ValueError("Page cannot be negative")
        del self.offset
        self._page = int(v)

    @page.deleter
    def page(self):
        self._page = None

    def __init__(self, limit=None, page=None, offset=None):
        self.paginate = False
        self._limit = limit
        self._offset = offset
        self._page = page

    def set(self, limit=None, page=None, offset=None):
        if limit is not None:
            self.limit = limit

        if page is not None:
            if offset:
                raise ValueError("Cannot pass in both offset and page")
            self.page = page

        if offset is not None:
            self.offset = offset

    def get(self):
        limit = self.limit_paginate if self.paginate else self.limit
        return (limit, self.offset)

    def __bool__(self):
        return self.limit > 0 or self.offset > 0

    def has(self):
        return bool(self)

    def has_limit(self):
        return self.limit > 0

    def has_more(self):
        """Returns True if the current bounds are set up to query one extra row in order
        to make pagination easier (to know if there should be a next link)

        :returns: boolean
        """
        return self.has_limit() and self.paginate

    def has_pages(self):
        return self.has_more()

    def is_paginated(self):
        return self.has_more()

    def __str__(self):
        return "limit: {}, offset: {}".format(self.limit, self.offset)

    def find_offset(self, i):
        """Given an index i, use the current offset and limit to find the correct
        offset i would be

        :param i: int, the index used to find the new offset
        :returns: int, the new offset
        """
        if i >= 0:
            offset = self.offset + i
            if self.has_limit():
                maximum_offset = self.limit + self.offset
                if offset > maximum_offset:
                    raise IndexError("Iterator index {} out of range".format(i))

        else:
            limit = self.limit
            offset = limit + i

            if offset < 0:
                raise IndexError("Iterator index {} out of range".format(i))

        return offset

    def find_more_index(self):
        #return self.offset + self.limit_paginate
        return self.offset + self.limit


class QueryField(object):
    @property
    def schema(self):
        return self.query.schema if self.query else None

    def __init__(self, query, field_name, field_val=None, **kwargs):
        self.query = query
        self.operator = kwargs.pop("operator", None)
        self.is_list = kwargs.pop("is_list", False)
        self.direction = kwargs.pop("direction", None) # 1 = ASC, -1 = DESC
        self.raw = kwargs.pop("raw", False)
        self.or_clause = False
        self.kwargs = kwargs

        if self.raw:
            self.name = field_name
            self.value = field_val

        else:
            self.set_name(field_name)
            self.set_value(field_val)

    def set_name(self, field_name):
        field_name, function_name = self.parse(field_name, self.schema)
        self.function_name = function_name
        self.name = field_name

    def set_value(self, field_val):
        if not isinstance(field_val, Query):
            if self.is_list:
                if field_val:
                    field_val = make_list(field_val)
                    for i in range(len(field_val)):
                        field_val[i] = self.iquery(field_val[i])

            else:
                field_val = self.iquery(field_val)

        self.value = field_val

    def iquery(self, field_val):
        query = self.query
        schema = self.schema
        if query and schema:
            schema_field = getattr(schema, self.name)
            if ref_class := schema_field.ref:
                if isinstance(field_val, ref_class):
                    field_val = field_val.pk

            field_val = schema_field.iquery(query, field_val)
        return field_val

    def parse(self, field_name, schema):
        function_name = ""
        if schema:
            function_name = ""
            m = re.match(r"^([^\(]+)\(([^\)]+)\)$", field_name)
            if m:
                function_name = m.group(1)
                field_name = m.group(2)

            try:
                field_name = schema.field_name(field_name)

            except AttributeError:
                field_name = schema.field_model_name(field_name)

        return field_name, function_name


class QueryFields(list):
    @property
    def fields(self):
        """Returns a dict of field_name: field_value"""
        ret = {}
        for f in self:
            ret[f.name] = f.value
        return ret

    def __init__(self):
        self.clear()

    def names(self):
        """Return all the field names in the order they were first seen"""
        ret = []
        seen = set()
        for f in self:
            if f.name not in seen:
                ret.append(f.name)
                seen.add(f.name)
        return ret

    def append(self, field):
        index = len(self)
        super().append(field)
        self.field_names[field.name].append(index)

    def __contains__(self, field_name):
        return field_name in self.field_names

    def __setitem__(self, *args, **kwargs):
        raise NotImplementedError()
    __delitem__ = __setitem__

    def clear(self):
        self.field_names = defaultdict(list)
        self.options = {}
        super().clear()


class Query(object):
    """Handle standard query creation and allow interface querying

    :example:
        q = Query(orm_class)
        q.eq_foo(1).desc_bar().limit(10).page(2).get()
        # SELECT
        #   *
        # FROM
        #   <orm_class.table_name>
        # WHERE
        #   foo = 1
        # ORDER BY bar DESC
        # LIMIT 10 OFFSET 20
    """
    field_class = QueryField
    fields_set_class = QueryFields
    fields_select_class = QueryFields
    fields_where_class = QueryFields
    fields_sort_class = QueryFields
    bounds_class = QueryBounds
    iterator_class = Iterator

    @property
    def interface(self):
        return self.orm_class.interface if self.orm_class else None

    @property
    def schema(self):
        if not self.orm_class: return None
        return self.orm_class.schema

    @property
    def schemas(self):
        """Find and return all the schemas that are needed for this query to complete
        successfully

        Another way to put this is all the schemas this query touches

        :returns: list, a list of Schema instances
        """
        schemas = []
        s = self.schema
        if s:
            schemas.append(s)

        for f in self.fields_where:
            if isinstance(f.value, Query):
                s = f.value.schema
                if s:
                    schemas.append(s)

        return schemas

    @property
    def OR(self):
        """Wraps left and right field statements with an OR clause, you can chain
        as many OR calls as you want to create an any length OR clause

        I don't love that I had to use OR instead of or but "or" is a reserved
        keyword and I thought OR was better than like _or, or_ or _or_

        :Example:
            self.eq_foo(1).OR.eq_foo(5).OR.eq_foo(10) # (foo=1 OR foo=5 OR foo=10)
        """
        self.fields_where[-1].or_clause = True
        return self

    @property
    def or_(self): return self.OR
    @property
    def _or(self): return self.OR

    @property
    def AND(self):
        """This is just here for completeness with .OR since, by default, any
        statements will be joined by AND"""
        self.fields_where[-1].or_clause = False
        return self

    @property
    def and_(self): return self.AND
    @property
    def _and(self): return self.AND

    def __init__(self, orm_class=None, **kwargs):
        # needed to use the db querying methods like get(), if you just want to build
        # a query then you don't need to bother passing this in
        self.orm_class = orm_class
        self.reset()

    def reset(self):
        self._ifilter = None
        self.fields_set = self.fields_set_class()
        self.fields_select = self.fields_select_class()
        self.fields_where = self.fields_where_class()
        self.fields_sort = self.fields_sort_class()
        self.bounds = self.bounds_class()
        self.compounds = []

    def ref(self, orm_classpath):
        """
        takes a classpath to allow query-ing from another Orm class

        the reason why it takes string paths is to avoid infinite recursion import 
        problems because an orm class from module A might have a ref from module B
        and sometimes it is handy to have module B be able to get the objects from
        module A that correspond to the object in module B, but you can't import
        module A into module B because module B already imports module A.

        :param orm_classpath: string|type, a full python class path (eg, foo.bar.Che) or
            an actual model.Orm python class
        return -- Query()
        """
        # split orm from module path
        if isinstance(orm_classpath, basestring):
            orm_module, orm_class = get_objects(orm_classpath)
        else:
            orm_class = orm_classpath

        return orm_class.query

    def ref_class(self, orm_classpath):
        return self.ref(orm_classpath)

    def __iter__(self):
        return self.get()

    def find_methods(self, method_name):
        """Given a method name like <OPERATOR>_<FIELD_NAME> or <FIELD_NAME>_<OPERATOR>,
        split those into <OPERATOR> and <FIELD_NAME> if there is an existing
        <OPERATOR>_field method that exists

        :example:
            self.gt_foo() # (<gt_field>, None, "foo")
            self.one_pk(value) # (<eq_field>, <one>, "pk")

        :returns: tuple, (<FIELD_METHOD>, <QUERY_METHOD>, <FIELD_NAME>)
        """
        # infinite recursion check, if a *_field method gets in here then it
        # doesn't exist
        if method_name.endswith("_field"):
            raise AttributeError(method_name)

        try:
            # check for <NAME>_<FIELD_NAME>
            name, field_name = method_name.split("_", 1)

        except ValueError:
            raise AttributeError("invalid potential method: {}".format(method_name))

        else:

            if not name:
                raise AttributeError('Could not resolve methods from {}"'.format(
                    method_name
                ))

            elif name in {"one", "value"}:
                field_method = self.eq_field
                query_method = getattr(self, name)

            elif name in {"get", "values", "all", "cursor"}:
                field_method = self.in_field
                query_method = getattr(self, name)

            elif name in {"raw"}:
                raise AttributeError(method_name)

            else:
                query_method = None
                field_method_name = "{}_field".format(name)
                field_method = getattr(self, field_method_name, None)
                if not field_method:
                    # let's try reversing the split, so <FIELD_NAME>_<NAME>
                    field_name, name = method_name.rsplit("_", 1)
                    field_method_name = "{}_field".format(name)
                    field_method = getattr(self, field_method_name, None)

                if not field_method:
                    raise AttributeError('No field method derived from {}'.format(
                        method_name
                    ))

        return field_method, query_method, field_name

    def __getattr__(self, method_name):
        field_method, query_method, field_name = self.find_methods(method_name)

        def callback(*args, **kwargs):
            #pout.v(args, kwargs, field_method, query_method, field_name)
            if field_method:
                ret = field_method(field_name, *args, **kwargs)

            if query_method:
                ret = query_method()

            return ret

        return callback

    def create_field(self, field_name, field_val=None, **kwargs):
        f = self.field_class(self, field_name, field_val, **kwargs)
        return f

    def create_iterator(self, query):
        if self.orm_class:
            iterator = self.orm_class.iterator_class(query)
        else:
            iterator = self.iterator_class(query)
        return iterator

    def append_compound(self, operator, queries, **kwargs):
        """Internal method used by .intersect(), .union(), and .difference()"""
        for i, query in enumerate(queries):
            if self.fields_select:
                if not query.fields_select:
                    if self.schema == query.schema:
                        query.select(*self.fields_select.names())

            else:
                if i == 0 and query.fields_select:
                    self.select(*query.fields_select.names())

        self.compounds.append((operator, queries))
        return self

    def append_operation(self, operator, field_name, field_val=None, **kwargs):
        kwargs["operator"] = operator
        f = self.create_field(field_name, field_val, **kwargs)
        self.fields_where.append(f)
        return self

    def append_sort(self, direction, field_name, field_val=None, **kwargs):
        """
        sort this query by field_name in directrion

        used to be named sort_field

        :param field_name: string, the field to sort on
        :param direction: integer, negative for DESC, positive for ASC
        :param field_val: list, the order the rows should be returned in
        """
        if direction > 0:
            direction = 1
        elif direction < 0:
            direction = -1
        else:
            raise ValueError("direction {} is undefined".format(direction))

        kwargs["direction"] = direction
        kwargs["is_list"] = True
        f = self.create_field(field_name, field_val, **kwargs)
        self.fields_sort.append(f)
        return self

    def distinct(self, *field_names):
        self.fields_select.options["distinct"] = True
        return self.select(*field_names)

    def select_field(self, field_name):
        """set a field to be selected, this is automatically called when you do select_FIELDNAME(...)"""
        if field_name == "*":
            self.fields_select.options["all"] = True
        else:
            field = self.create_field(field_name)
            self.fields_select.append(field)
        return self

    def select(self, *field_names):
        """set multiple fields to be selected"""
        for field_name in make_list(field_names):
            self.select_field(field_name)
        return self

    def set_field(self, field_name, field_val):
        """
        set a field into .fields_set attribute

        In insert/update queries, these are the fields that will be inserted/updated into the db
        """
        field = self.create_field(field_name, field_val)
        self.fields_set.append(field)
        return self

    def set(self, fields=None, **fields_kwargs):
        """
        completely replaces the current .fields with fields and fields_kwargs combined
        """
        fields = make_dict(fields, fields_kwargs)
        for field_name, field_val in fields.items():
            self.set_field(field_name, field_val)
        return self

    def intersect(self, *queries, **kwargs):
        """Intersect a set of queries. Returns rows that are common in all the 
        queries.

        Compound Query instances will only return values, not Orm instances

        :Example:
            Query().intersect(
                Query().select_foo(),
                Query().select_foo(),
            ).all()

        https://www.sqlite.org/syntax/compound-operator.html

        :param *queries: two or more queries, they should all select the same
            field types, the first query sets the name of the return columns
        """
        return self.append_compound("intersect", queries, **kwargs)

    def union(self, *queries, **kwargs):
        """Return all the rows from all the queries

        See .intersect() since this works very similar
        """
        return self.append_compound("union", queries, **kwargs)

    def difference(self, *queries, **kwargs):
        """Return everything in the first query that isn't in any of the other
        queries.

        Return the rows with rows in the first query that are not in the other
        queries

        This is named .difference() instead of .except() because except is a python
        keyword.

        See .intersect() since this works very similar
        """
        return self.append_compound("except", queries, **kwargs)

    def is_field(self, field_name, field_val=None, **field_kwargs):
        return self.eq_field(field_name, field_val, **field_kwargs)
    def eq_field(self, field_name, field_val=None, **field_kwargs):
        return self.append_operation("eq", field_name, field_val, **field_kwargs)

    def not_field(self, field_name, field_val=None, **field_kwargs):
        return self.ne_field(field_name, field_val, **field_kwargs)
    def ne_field(self, field_name, field_val=None, **field_kwargs):
        return self.append_operation("ne", field_name, field_val, **field_kwargs)

    def between_field(self, field_name, low, high):
        self.gte_field(field_name, low)
        self.lte_field(field_name, high)
        return self

    def lte_field(self, field_name, field_val=None, **field_kwargs):
        return self.append_operation("lte", field_name, field_val, **field_kwargs)

    def lt_field(self, field_name, field_val=None, **field_kwargs):
        return self.append_operation("lt", field_name, field_val, **field_kwargs)

    def gte_field(self, field_name, field_val=None, **field_kwargs):
        return self.append_operation("gte", field_name, field_val, **field_kwargs)

    def gt_field(self, field_name, field_val=None, **field_kwargs):
        return self.append_operation("gt", field_name, field_val, **field_kwargs)

    def in_field(self, field_name, field_val=None, **field_kwargs):
        """
        :param field_val: list, a list of field_val values
        """
        field_kwargs["is_list"] = True
        return self.append_operation("in", field_name, field_val, **field_kwargs)

    def nin_field(self, field_name, field_val=None, **field_kwargs):
        """
        :param field_val: list, a list of field_val values
        """
        field_kwargs["is_list"] = True
        return self.append_operation("nin", field_name, field_val, **field_kwargs)

    def startswith_field(self, field_name, field_val, **field_kwargs):
        return self.like_field(field_name, "{}%".format(field_val), **field_kwargs)

    def endswith_field(self, field_name, field_val, **field_kwargs):
        return self.like_field(field_name, "%{}".format(field_val), **field_kwargs)

    def contains_field(self, field_name, field_val, **field_kwargs):
        return self.like_field(field_name, "%{}%".format(field_val), **field_kwargs)

    def like_field(self, field_name, field_val, **field_kwargs):
        """Perform a field_name LIKE field_val query

        :param field_name: string, the field we are filtering on
        :param field_val: string, the like query: %val, %val%, val%
        :returns: self, for fluid interface
        """
        if not field_val:
            raise ValueError("Cannot LIKE nothing")
        return self.append_operation("like", field_name, field_val, **field_kwargs)

    def nlike_field(self, field_name, field_val, **field_kwargs):
        """Perform a field_name NOT LIKE field_val query

        :param field_name: string, the field we are filtering on
        :param field_val: string, the like query: %val, %val%, val%
        :returns: self, for fluid interface
        """
        if not field_val:
            raise ValueError("Cannot NOT LIKE nothing")
        return self.append_operation("nlike", field_name, field_val, **field_kwargs)

    def raw_field(self, field_name, *field_vals, **field_kwargs):
        field_kwargs["raw"] = True
        return self.append_operation("raw", field_name, field_vals, **field_kwargs)

    def asc(self, *field_names):
        for field_name in field_names:
            self.asc_field(field_name)
        return self

    def asc_field(self, field_name, field_val=None):
        return self.append_sort(1, field_name, field_val)

    def desc(self, *field_names):
        for field_name in field_names:
            self.desc_field(field_name)
        return self

    def desc_field(self, field_name, field_val=None):
        return self.append_sort(-1, field_name, field_val)

    def ifilter(self, predicate):
        """Set the predicate (callback) for an iterator returned from this instance

        https://docs.python.org/2/library/itertools.html#itertools.ifilter

        :param predicate: callable, this will be passed to the iterator in the
            .create_iterator method
        :returns: self, for fluid interface
        """
        self._ifilter = predicate
        return self

    def filter(self, predicate):
        return self.ifilter(predicate)

    def limit(self, limit):
        self.bounds.limit = limit
        return self

    def offset(self, offset):
        self.bounds.offset = offset
        return self

    def page(self, page):
        self.bounds.page = page
        return self

    def cursor(self):
        """Used by the Iterator to actually query the db"""
        return self.execute('get', cursor_result=True)

    def get(self):
        """
        get results from the db

        :returns: Iterator
        """
        self.bounds.paginate = True
        return self.create_iterator(self)

    def all(self):
        return self.get()

    def values(self):
        if not self.fields_select:
            raise ValueError("No selected fields")
        return self.get()

    def one(self):
        """get one row from the db"""
        self.limit(1)
        self.bounds.paginate = False
        it = self.create_iterator(self)
        try:
            ret = it.next()
        except StopIteration:
            ret = None
        return ret

    def value(self):
        """convenience method to just get one value or tuple of values for the query"""
        if not self.fields_select:
            raise ValueError("no selected fields")
        return self.one()

    def count(self):
        """return the count of the criteria"""

        # sorting shouldn't matter for a count query
        fields_sort = self.fields_sort
        self.fields_sort = self.fields_sort_class()

        # more than one selected field will cause the count query to error out
        fields_select = self.fields_select
        self.fields_select = self.fields_select_class()

        # setting bounds causes count(*) to return 0 in both Postgres and SQLite
        bounds = self.bounds
        self.bounds = self.bounds_class()

        ret = self.execute('count')

        # restore previous values now that count is done
        self.fields_sort = fields_sort
        self.fields_select = fields_select
        self.bounds = bounds

        # now we are going to compensate for the bounds being set
        if self.bounds:
            offset = self.bounds.offset
            ret -= offset
            if self.bounds.has_limit():
                limit = self.bounds.limit
                if ret > limit:
                    ret = limit

        return ret

    def has(self):
        """returns true if there is atleast one row in the db matching the query, False otherwise"""
        v = self.one()
        return True if v else False

    def exists(self):
        return self.has()

    def insert(self, **kwargs):
        """persist the .fields"""
        return self.interface.insert(self.schema, self.fields_set.fields, **kwargs)

    def update(self, **kwargs):
        """persist the .fields using .fields_where"""
        return self.interface.update(
            self.schema,
            self.fields_set.fields,
            self,
            **kwargs
        )

    def upsert(self, conflict_field_names=None, **kwargs):
        """persist the .fields"""
        insert_fields = self.fields_set.fields
        update_fields = dict(insert_fields)

        if not conflict_field_names:
            conflict_field_names = self.schema.pk_names

        for field_name in conflict_field_names:
            update_fields.pop(field_name, None)

        return self.interface.upsert(
            self.schema,
            insert_fields,
            update_fields,
            conflict_field_names,
            **kwargs
        )

    def delete(self, **kwargs):
        """remove fields matching the where criteria"""
        return self.execute('delete', **kwargs)

    def render(self, **kwargs):
        """Render the query

        :returns: string, the rendered query, this is not assured to be a valid query
            but is handy for quickly debugging what the query roughly looks like
        """
        return self.interface.render(self.schema, self, **kwargs)

    def raw(self, query_str, *query_args, **query_options):
        """
        use the interface.query() method to pass in your own raw query without
        any processing

        NOTE -- This will allow you to make any raw query and will usually return
        raw results, it won't wrap those results in a prom.Orm iterator instance
        like other methods like .all() and .get()

        query_str -- string -- the raw query for whatever the backend interface is
        query_args -- list -- if you have named parameters, you can pass in the values
        **query_options -- dict -- key:val options for the backend, these are very backend specific
        return -- mixed -- depends on the backend and the type of query
        """
        i = self.interface
        return i.raw(query_str, *query_args, **query_options)

    def execute(self, method_name, **kwargs):
        i = self.interface
        s = self.schema
        return getattr(i, method_name)(s, self, **kwargs)

    def copy(self):
        """nice handy wrapper around the deepcopy"""
        return copy.deepcopy(self)

    def __deepcopy__(self, memodict={}):
        instance = type(self)(self.orm_class)
        ignore_keys = set(["_interface", "interface"])
        for key, val in self.__dict__.items():
            if key not in ignore_keys:
                setattr(instance, key, copy.deepcopy(val, memodict))
        return instance

    def __str__(self):
        try:
            return self.render()

        except AttributeError:
            return super().__str__()




# -*- coding: utf-8 -*-
import datetime
import functools
import logging
import uuid

from testdata.base import TestData
from datatypes import (
    OrderedSubclasses,
    ReflectModule,
)

from ..model import Orm
from ..interface import get_interfaces
from ..exception import UniqueError


logger = logging.getLogger(__name__)


class ModelData(TestData):
    """Provides testdata hooks for the projects models, testdata is a python module
    for generating random data for testing

    https://github.com/Jaymon/testdata

    This uses a lot of magic and I'm not sure the best way to describe it, basically,
    this will create get_*, create_*, and get_*_fields methods for any Orm subclass
    you have loaded into memory. The * will correspond to Orm.model_name and 
    Orm.models_name.

    Because it only creates those methods for Orm classes loaded
    into memory you need to import the modules for the classes you want to test

    Maybe the best way to understand what this does is by example.

    :Example:
        from prom import Orm

        class Foobar(Orm):
            pass

        import testdata
        from prom.extras.testdata import ModelData

        o = testdata.get_foobar()
        print(o.pk) # None

        o = testdata.create_foobar()
        print(o.pk) # pk will be set because .save() was called

        fields = testdata.get_foobar_fields()
        print(fields) # dict that can be used create a Foobar(fields) instance

        os = testdata.get_foobars(foobar_count=2) # will return a list of 2 Foobar instances

        os = testdata.create_foobars(foobar_count=2) # will return a list of 2 created Foobar instances


    To customize a method, you can extend this class and define the method. The
    signature for any overridden methods is (self, orm_class, **kwargs),
    and you will usually just want to override the get_<MODELS-NAME>_fields class
    since that is the one the get_* and create_* methods use to populate the orms

    All the signatures of all the methods need to be the same because they are
    mixed and matched, so if you have a `Foobar` Orm class and override the
    get_foobar_fields testdata method, then get_foobar() would call get_foobar_fields
    and get_orm(Foobar, **kwargs) would also call get_foobar_fields. So all of these
    methods can be mixed and matched

    :Example:
        from prom.extras.testdata import ModelData

        class MyModelData(ModelData):
            def get_foobar_fields(self, orm_class, **kwargs):
                return super().get_orm_fields(orm_class, **kwargs)

    If you'd like to customize the testdata method names, you can set the Orm.model_name
    or the Orm.models_name class properties

    :Example:
        from prom import Orm

        class Foobar(Orm):
            model_name = "foo_bar"
            models_name = "foo_bars"
    """
    model_cache = {}
    """Caches the found models for a given model name"""

    def _orm_classes(self):
        """Iterate through a list of orms that should be injected into testdata

        by default, we want to ignore any orm classes that are defined in this
        library since they are by definition base classes

        :returns: generator, each orm class
        """
        orm_classes = OrderedSubclasses(classes=Orm.orm_classes.values())
        module_name = ReflectModule(__name__).modroot

        for orm_class in orm_classes.edges():
            if module_name not in orm_class.__module__:
                yield orm_class

    def _find_method(self, method_name, default_method=None, **kwargs):
        """Find the method, this will return the user defined method or the default
        method if no user defined method exists

        :param method_name: str, the method name
        :param default_method: callable, the fallback method
        :param **kwargs:
        :returns: callable
        """
        method = getattr(self, method_name, None)
        if method:
            return method_name, method

        else:
            return method_name, default_method

    def _get_method(self, orm_class, **kwargs):
        """Return the .get_orm() type method for orm_class"""
        method_name = f"get_{orm_class.model_name}"
        kwargs.setdefault("default_method", self.get_orm)
        return self._find_method(method_name, **kwargs)

    def _get(self, orm_class, **kwargs):
        """Internal dispatcher method for get_orm, this will first try and find
        a get_<ORM-NAME> method and fallback to get_orm"""
        method_name, method = self._get_method(orm_class)
        return self._run_method(method_name, method, orm_class, **kwargs)

    def _gets_count(self, orm_class, **kwargs):
        """Find how many orm_class should be created

        :param orm_class: Orm
        :param **kwargs: this will be checked for the correct *_count or count
            value
        :returns: int, how many of orm_class is wanted
        """
        return kwargs.get(
            f"{orm_class.model_name}_count",
            kwargs.get(
                f"{orm_class.models_name}_count",
                kwargs.get(
                    "count",
                    1
                )
            )
        )

    def _gets_method(self, orm_class, **kwargs):
        """Return the .get_orms() type method for orm_class"""
        method_name = f"get_{orm_class.models_name}"
        kwargs.setdefault("default_method", self.get_orms)
        return self._find_method(method_name, **kwargs)

    def _gets(self, orm_class, **kwargs):
        """Internal dispatcher method for get_orms, this will first try and find
        a get_<ORM-MODELS-NAME> method and fallback to get_orms"""
        method_name, method = self._gets_method(orm_class)
        return self._run_method(method_name, method, orm_class, **kwargs)

    def _create_method(self, orm_class, **kwargs):
        """Return the .create_orm() type method for orm_class"""
        method_name = f"create_{orm_class.model_name}"
        kwargs.setdefault("default_method", self.create_orm)
        return self._find_method(method_name, **kwargs)

    def _create(self, orm_class, **kwargs):
        """Internal dispatcher method for create_orm, this will first try and find
        a create_<ORM-NAME> method and fallback to create_orm"""
        method_name, method = self._create_method(orm_class)
        return self._run_method(method_name, method, orm_class, **kwargs)

    def _creates_method(self, orm_class, **kwargs):
        """Return the .create_orms() type method for orm_class"""
        method_name = f"create_{orm_class.models_name}"
        kwargs.setdefault("default_method", self.create_orms)
        return self._find_method(method_name, **kwargs)

    def _creates(self, orm_class, **kwargs):
        """Internal dispatcher method for create_orms, this will first try and find
        a create_<ORM-MODELS-NAME> method and fallback to create_orm"""
        method_name, method = self._creates_method(orm_class)
        return self._run_method(method_name, method, orm_class, **kwargs)

    def _fields_method(self, orm_class, **kwargs):
        """Return the .get_orm_fields() type method for orm_class"""
        method_name = f"get_{orm_class.model_name}_fields"
        kwargs.setdefault("default_method", self.get_orm_fields)
        return self._find_method(method_name, **kwargs)

    def _fields(self, orm_class, **kwargs):
        """Internal dispatcher method for get_orm_fields, this will first try and find
        a get_<ORM-NAME>_fields method and fallback to get_orm_fields"""
        method_name, method = self._fields_method(orm_class)
        return self._run_method(method_name, method, orm_class, **kwargs)

    def _run_method(self, method_name, method, orm_class, **kwargs):
        kwargs.setdefault("orm_class", orm_class)
        logger.debug(
            "Running {} as {} for orm_class {}".format(
                method.__name__,
                method_name,
                orm_class.__name__,
            )
        )
        return method(**kwargs)

    def _find_attr(self, method_name):
        """Internal method used by self.__getattr__ to find the orm and create a
        wrapper method to call

        NOTE -- this method must raise AttributeError on expected errors

        :param method_name: str, the method name that contains the model class
            we're ultimately looking for
        :returns: callable, the method that should be ran for the passed in
            method_name
        """
        logger.debug(
            f"Finding {self.__class__.__name__}.{method_name} method"
        )

        try:
            # check for <NAME>_<MODEL_NAME>
            name, model_name = method_name.split("_", 1)

            if model_name.endswith("_fields"):
                name = "fields"
                model_name, _ = model_name.rsplit("_", 1)

        except ValueError as e:
            raise AttributeError(
                f"invalid potential method: {method_name}"
            ) from e

        else:
            try:
                orm_class = self.get_orm_class(model_name)

            except ValueError as e:
                raise AttributeError(
                    f"Could not find orm class from {method_name}"
                ) from e

            else:
                orm_method = None
                if name == "get":
                    if orm_class.model_name == model_name:
                        orm_method = self.get_orm

                    else:
                        orm_method = self.get_orms

                elif name == "create":
                    if orm_class.model_name == model_name:
                        orm_method = self.create_orm

                    else:
                        orm_method = self.create_orms

                elif name == "fields":
                    orm_method = self.get_orm_fields

                if orm_method:
                    logger.debug(
                        f"Found {orm_method.__name__} for {orm_class.__name__}"
                    )

                    def method(**kwargs):
                        # https://github.com/Jaymon/prom/issues/166
                        # we want to override the passed in orm_class if it
                        # doesn't match our found orm class because this has
                        # most likely been called internally by another magic
                        # method that just passed kwargs
                        kwargs.setdefault("orm_class", orm_class)
                        if not isinstance(orm_class, kwargs["orm_class"]):
                            kwargs["orm_class"] = orm_class

                        return orm_method(**kwargs)

                    # could also use functools.wraps here on method instead of
                    # just setting the name, but I like that it explicitely
                    # says it is wrapped in the name instead of just
                    # transparantly using orm_method's name
                    method.__name__ = f"wrapped_{orm_method.__name__}"
                    return method

        raise AttributeError(f"Could not find an orm matching {method_name}")

    def __getattr__(self, method_name):
        try:
            return self._find_attr(method_name)

        except AttributeError:
            return super().__getattr__(method_name)

    def unsafe_delete_db(self):
        """This will delete all the tables from the db

        It relies on a method .ensure_safe_env being defined in the project code
        """
        if self.ensure_safe_env(): # this method needs to be defined in project code
            for inter in get_interfaces().values():
                inter.unsafe_delete_tables()

    def unsafe_install_orms(self, modulepaths):
        """Go through and install all the Orm subclasses found in the passed in
        module paths

        :param modulepaths: Sequence[str], a list of modpaths (eg ["foo.bar", "che"])
        """
        # import the module paths to load the Orms into memory
        for modulepath in modulepaths:
            rm = ReflectModule(modulepath)
            m = rm.module() 

        # now go through all the orm classes that have been loaded and install them
        seen_table_names = set()
        for orm_class in self._orm_classes():
            for s in orm_class.schema.schemas:
                if s.table_name not in seen_table_names:
                    s.orm_class.install()
                    seen_table_names.add(s.table_name)

    def assure_orm_field_names(self, orm_class, **kwargs):
        """Field instances can have aliases, in order to allow you to pass in aliases,
        this will go through kwargs and normalize the field names

        :Example:
            class Foobar(Orm):
                che = Field(str, aliases=["baz"]

            kwargs = testdata.assure_orm_field_names(Foobar, {"baz": "1"})
            print(kwargs["che"]) # "1"

        :param orm_class: Orm
        :param **kwargs: the fields where keys will be normalized to field names
            in orm_class.schema
        :returns: dict, the normalized kwargs
        """
        schema = orm_class.schema
        # normalize passed in field names to make sure we correctly find the field's
        # value if it exists
        for field_name in list(kwargs.keys()):
            if schema.has_field(field_name):
                kwargs.setdefault(schema.field_name(field_name), kwargs.pop(field_name))

        return kwargs

    def assure_ref_field_names(self, orm_class, ref_class, **kwargs):
        """Make sure the kwargs destined for orm_class don't impact the kwargs
        that will be used to create a ref_class instance

        Sometimes, we have multiple Orm classes that have the same field names and
        when those fields get set for orm_class they can cause ref_class to fail
        so we want to strip those common field names out

        :param orm_class: Orm, the main orm class
        :param ref_class: Orm, the Orm that orm_class references in some way
        :param **kwargs: the fields that were passed in for orm_class, if these
            fields are also in ref_class they will be stripped
        :returns: dict, the kwargs suitable to be used to create an instance of
            ref_class
        """
        ref_kwargs = {}
        orm_fields = orm_class.schema.fields
        ref_fields = ref_class.schema.fields
        for field_name, field_value in kwargs.items():
            if (field_name not in orm_fields) or (field_name not in ref_fields):
                ref_kwargs[field_name] = field_value

        return ref_kwargs

    def assure_orm_refs(self, orm_class, assure_orm_class=None, **kwargs):
        """When creating an orm, they will often need foreign key values, this will
        go through any of the foreign key ref fields and create a foreign key if
        it wasn't included.

        This is a recursive method, when it finds a ref_class it will assure that
        ref's foreign keys before handling ref, this way all references all the
        way down the stack can be generated for the top level so the same refs
        can propogate all the way down the dependency chain

        :param orm_class: Orm
        :param assure_orm_class: Orm|None, the orm that is getting its references
            assured. If this is None then orm_class is considered the orm that is
            being checked
        :param **kwargs: orm_class's actual field name value will be checked and
            the ref's orm_class.model_name will be checked
            * ignore_refs: bool, default False, if True then refs will not be checked,
                passed in refs will still be set
            * require_fields: bool, default True, if True then create missing refs,
                if False, then refs won't be created and so if they are missing
                their fields will not be populated
        :returns: dict, the kwargs with ref field_name and ref orm_class.model_name
            will be included
        """
        logger.debug(f"Assuring orm refs for orm_class {orm_class.__name__}")
        kwargs = self.assure_orm_field_names(orm_class, **kwargs)

        # if assure class isn't passed in then we assume the passed in orm_class
        # is the class to be assured and all recursive calls will now have it
        # set
        if assure_orm_class is None:
            assure_orm_class = orm_class

        ignore_refs = kwargs.get("ignore_refs", False)
        require_fields = kwargs.get("require_fields", True)
        for field_name, field in orm_class.schema.fields.items():
            if ref_class := field.ref:
                ref_field_name = ref_class.model_name

                if field_name in kwargs:
                    if ref_field_name not in kwargs:
                        kwargs[ref_field_name] = ref_class.query.eq_pk(
                            kwargs[field_name]
                        ).one()

                else:
                    if ref_field_name in kwargs:
                        kwargs[field_name] = kwargs[ref_field_name].pk

                    else:
                        if not ignore_refs:
                            if require_fields or field.is_required() or self.yes():
                                # handle all ref_class's refs before we handle
                                # ref_class
                                kwargs.update(self.assure_orm_refs(
                                    ref_class,
                                    assure_orm_class,
                                    **kwargs
                                ))

                                kwargs[ref_field_name] = self._create(
                                    ref_class,
                                    **self.assure_ref_field_names(
                                        assure_orm_class,
                                        ref_class,
                                        **kwargs
                                    )
                                )
                                kwargs[field_name] = kwargs[ref_field_name].pk

        return kwargs

    def create_orm(self, orm_class, **kwargs):
        """create an instance of the orm and save it into the db

        :param orm_class: Orm
        :param **kwargs:
        :returns: Orm, the orm saved into the db
        """
        kwargs.setdefault("ignore_refs", False)
        instance = self._get(orm_class, **kwargs)
        try:
            instance.save(nest=True)

        except UniqueError as e:
            logger.warning(" ".join([
                f"Creating {orm_class.__name__} failed because it exists.",
                "Fetching the existing instance without updating it",
            ]))
            instance = instance.requery()

        for k in list(kwargs.keys()):
            # We want to create any orms with FK references to orm_class if  counts
            # were passed in
            if k.endswith("_count"):
                k_model_name = k[0:-6]
                k_orm_class = self.get_orm_class(k_model_name)
                if k_orm_class:
                    logger.debug("Creating {} {} instances tied to {} instance".format(
                        kwargs[k],
                        k_orm_class.__name__,
                        orm_class.__name__,
                    ))
                    kwargs.setdefault(instance.model_name, instance)
                    self.create_orms(k_orm_class, **kwargs)

        return instance

    def create_orms(self, orm_class, **kwargs):
        """create instances of the orm and save it into the db

        :param orm_class: Orm
        :param **kwargs:
            count: int, how many instances you want
            <MODEL_NAME>_count: int, alias for count
        :returns: list, a list of Orm instances that have all been saved into the db
        """
        kwargs.setdefault("ignore_refs", False)
        instances = self._gets(orm_class, **kwargs)
        for instance in instances:
            instance.save(nest=True)
        return instances

    def get_orm_class(self, model_name, **kwargs):
        """get the orm class found at model_name

        :param model_name: str, the name of the orm class
        :returns: Orm, the orm_class.model_name that matches model_name
        """
        if model_name in self.model_cache:
            return self.model_cache[model_name]

        elif model_name in Orm.orm_classes:
            orm_class = Orm.orm_classes[model_name]
            self.model_cache[model_name] = orm_class
            return orm_class

        else:
            for oc in self._orm_classes():
                if model_name in set([oc.model_name, oc.models_name]):
                    orm_class = oc
                    self.model_cache[model_name] = orm_class
                    return orm_class

            for orm_classpath, orm_class in Orm.orm_classes.items():
                if orm_classpath.endswith(f":{model_name}"):
                    self.model_cache[model_name] = orm_class
                    return orm_class

        raise ValueError(f"could not find an orm_class for {model_name}")

    def get_orm(self, orm_class, **kwargs):
        """get an instance of the orm but don't save it into the db

        :param orm_class: Orm
        :param **kwargs:
        :returns: Orm
        """
        kwargs.setdefault("ignore_refs", True)
        instance = kwargs.get(orm_class.model_name, None)
        if not instance:
            fields = self._fields(orm_class, **kwargs)
            instance = orm_class(fields)

        return instance

    def get_orms(self, orm_class, **kwargs):
        """get instances of the orm

        :param orm_class: Orm
        :param **kwargs:
            * count: int, how many instances you want
            * <MODEL_NAME>_count: int, alias for count
            * related_refs: bool, default True, all the orms will have the same
                foreign key references
        :returns: list, a list of Orm instances
        """
        ret = []
        if kwargs.get("related_refs", True):
            # because we need related refs, we will need to create refs if they
            # don't exist
            kwargs.setdefault("ignore_refs", False)
            kwargs = self.assure_orm_refs(orm_class, **kwargs)

        count = self._gets_count(orm_class, **kwargs)
        for _ in range(count):
            ret.append(self._get(orm_class, **kwargs))

        return ret

    def get_orm_fields(self, orm_class, **kwargs):
        """Get the fields of an orm_class

        :param orm_class: Orm
        :param **kwargs: the fields found in orm_class.schema
            * require_fields: bool, default True, this will require that all fields
                have values even if they aren't required, this does not apply to
                foreign key references
            * field_callbacks: dict, the key is the field name and the value is
                a callable that can take self
        :returns: dict
        """
        ret = {}
        schema = orm_class.schema
        kwargs.setdefault("ignore_refs", True)
        kwargs = self.assure_orm_refs(orm_class, **kwargs)
        require_fields = kwargs.get("require_fields", True)
        field_callbacks = kwargs.get("field_callbacks", {})

        for field_name, field in schema.fields.items():
            if field_name in kwargs:
                # this value was passed in so we don't need to do anything
                ret[field_name] = kwargs[field_name]

            elif field.is_auto():
                # db will handle any auto-generating fields
                pass

            elif field.is_pk():
                # primary key isn't auto-generating and wasn't passed in, so
                # we'll cross our fingers and hope it will be taken care of
                # somewhere else
                pass

            elif field.is_ref():
                # foreign keys are handled in .assure_orm_refs
                pass

            else:
                if require_fields or field.is_required() or self.yes():

                    field_type = field.interface_type

                    if field.choices:
                        ret[field_name] = self.choice(field.choices)

                    else:
                        if cb := getattr(field, "testdata", field_callbacks.get(field_name, None)):
                            ret[field_name] = cb(self)

                        else:
                            if issubclass(field_type, bool):
                                ret[field_name] = bool(self.yes())

                            elif issubclass(field_type, int):
                                size_info = field.size_info()
                                ret[field_name] = self.get_posint(size_info["size"])

                            elif issubclass(field_type, str):
                                size_info = field.size_info()
                                if "bounds" in size_info:
                                    ret[field_name] = self.get_words(
                                        min_size=size_info["bounds"][0],
                                        max_size=size_info["bounds"][1],
                                    )

                                else:
                                    ret[field_name] = self.get_words()

                            elif issubclass(field_type, dict):
                                ret[field_name] = self.get_dict()

                            elif issubclass(field_type, float):
                                size_info = field.size_info()
                                ret[field_name] = self.get_posfloat(size_info["size"])

                            elif issubclass(field_type, datetime.datetime):
                                ret[field_name] = self.get_past_datetime()

                            elif issubclass(field_type, datetime.date):
                                ret[field_name] = self.get_past_date()

                            elif issubclass(field_type, uuid.UUID):
                                ret[field_name] = str(uuid.uuid4())

                            else:
                                raise ValueError(f"Not sure what to do with {field.type}")

        return ret




# -*- coding: utf-8 -*-
from __future__ import unicode_literals, division, print_function, absolute_import
import os
import datetime
import decimal
import logging
import uuid
from functools import cached_property

from ..query import Query
from ..exception import (
    TableError,
    FieldError,
    UniqueError,
    PlaceholderError,
)

from ..compat import *
from ..utils import make_list
from .base import Connection, Interface


logger = logging.getLogger(__name__)


class SQLConnection(Connection):
    """
    https://docs.python.org/3.9/library/sqlite3.html#sqlite3-controlling-transactions
    https://www.sqlite.org/lockingv3.html
    https://www.sqlite.org/lang_transaction.html
    """
    def _transaction_start(self):
        self._execute("BEGIN")

    def _transaction_starting(self, tx):
        # http://www.postgresql.org/docs/9.2/static/sql-savepoint.html
        self._execute("SAVEPOINT {}".format(
            self.interface.render_field_name_sql(tx["name"])
        ))

    def _transaction_stop(self):
        """
        http://initd.org/psycopg/docs/usage.html#transactions-control
        https://news.ycombinator.com/item?id=4269241
        """
        self._execute("COMMIT")

    def _transaction_stopping(self, tx):
        self._execute("RELEASE {}".format(
            self.interface.render_field_name_sql(tx["name"])
        ))

    def _transaction_fail(self):
        self._execute("ROLLBACK")

    def _transaction_failing(self, tx):
        # http://www.postgresql.org/docs/9.2/static/sql-rollback-to.html
        self._execute("ROLLBACK TO SAVEPOINT {}".format(
            self.interface.render_field_name_sql(tx["name"])
        ))

    def _execute(self, query_str):
        self.log_info(f"{self} - {query_str}")
        cur = self.cursor()
        cur.execute(query_str)
        cur.close()


class SQLInterfaceABC(Interface):
    """SQL database interfaces should extend SQLInterface and implement all these
    methods in this class and all the methods in InterfaceABC"""
    @property
    def LIMIT_NONE(self):
        """When an offset is set but not a limit, this is the value that will be
        put into the LIMIT part of the query

        :returns: str|int|None
        """
        raise NotImplementedError("this property should be set in any children class")

    def get_paramstyle(self):
        """Returns the paramstyle that is used by self.PLACEHOLDER to decide what
        val placeholder to use when building queries. This would also be the value
        to use when calling .raw().

        The dbapi 2.0 spec requires this to be set

        https://peps.python.org/pep-0249/#paramstyle

        :returns: str, something like "qmark" (for ? placeholders) or "pyformat"
            (for %s placeholders)
        """
        raise NotImplemented()

    def render_date_field_sql(self, field_name, field_kwargs, symbol):
        raise NotImplemented()

    def render_sort_field_sql(self, field_name, field_vals, sort_dir_str):
        """normalize the sort string

        return -- tuple -- field_sort_str, field_sort_args"""
        raise NotImplemented()

    def render_nolimit_sql(limit, **kwargs):
        raise NotImplemented()

    def render_datatype_datetime_sql(self, field_name, field, **kwargs):
        raise NotImplementedError()

    def render_datatype_dict_sql(self, field_name, field, **kwargs):
        raise NotImplementedError()

    def render_datatype_uuid_sql(self, field_name, field, **kwargs):
        raise NotImplementedError()


class SQLInterface(SQLInterfaceABC):
    """Generic base class for all SQL derived interfaces"""
    @cached_property
    def PLACEHOLDER(self):
        """What placeholder value this interface uses when building queries.

        This uses .get_paramstyle() to decide what placeholder value to use

        https://www.psycopg.org/docs/usage.html#passing-parameters-to-sql-queries

        NOTE -- It looks like both SQLite and Postgres support "named" so if you
            want to use .raw() queries that will work in both interface I would use
            named parameters

        :returns: str, usually something like "?" or "%s"
        """
        paramstyle = self.get_paramstyle()

        if paramstyle == "qmark":
            return "?"

        elif paramstyle == "format":
            return "%s"

        elif paramstyle == "pyformat":
            return "%s"

        elif paramstyle == "numeric":
            raise NotImplementedError("These are :1 :2 :3")

        elif paramstyle == "named":
            raise NotImplementedError("These are :name")

        else:
            raise NotImplementedError(f"Unknown paramstyle {paramstyle}")

    def _set_table(self, schema, **kwargs):
        """
        http://sqlite.org/lang_createtable.html
        http://www.postgresql.org/docs/9.1/static/sql-createtable.html
        http://www.postgresql.org/docs/8.1/static/datatype.html
        http://pythonhosted.org/psycopg2/usage.html#adaptation-of-python-values-to-sql-types
        """
        query_str = []
        query_str.append("CREATE TABLE IF NOT EXISTS {} (".format(self.render_table_name_sql(schema)))

        query_fields = []
        for field_name, field in schema.fields.items():
            query_fields.append('  {}'.format(self.render_datatype_sql(field_name, field)))

        query_str.append(",\n".join(query_fields))
        query_str.append(')')
        query_str = "\n".join(query_str)
        self._raw(query_str, ignore_result=True, **kwargs)

    def _set_index(self, schema, name, field_names, **kwargs):
        """
        NOTE -- we set the index name using <table_name>_<name> format since indexes have to have
        a globally unique name in postgres

        * https://www.sqlite.org/lang_createindex.html
        * https://www.postgresql.org/docs/14/sql-createindex.html - "IF NOT EXISTS support
        was added around 9.5
        """
        query_str = 'CREATE {}INDEX IF NOT EXISTS {} ON {} ({})'.format(
            'UNIQUE ' if kwargs.get('unique', False) else '',
            self.render_field_name_sql(f"{schema}_{name}"),
            self.render_table_name_sql(schema),
            ', '.join(map(self.render_field_name_sql, field_names))
        )

        return self._raw(query_str, ignore_result=True, **kwargs)

    def _insert(self, schema, fields, **kwargs):
        pk_names = schema.pk_names
        kwargs.setdefault("ignore_return_clause", len(pk_names) == 0)
        kwargs.setdefault("ignore_result", len(pk_names) == 0)

        query_str, query_args = self.render_insert_sql(
            schema,
            fields,
            **kwargs,
        )

        r = self._raw(query_str, *query_args, **kwargs)
        if r and pk_names:
            if len(pk_names) > 1:
                r = r[0]
            else:
                r = r[0][pk_names[0]]
        return r

    def _update(self, schema, fields, query, **kwargs):
        query_str, query_args = self.render_update_sql(
            schema,
            fields,
            query=query,
            **kwargs,
        )

        return self._raw(query_str, *query_args, count_result=True, **kwargs)

    def _upsert(self, schema, insert_fields, update_fields, conflict_field_names, **kwargs):
        """
        https://www.sqlite.org/lang_UPSERT.html
        """
        if not conflict_field_names:
            raise ValueError(f"Upsert is missing conflict fields for {schema}")

        for field_name in conflict_field_names:
            # conflict fields need to be in the insert fields
            if field_name not in insert_fields:
                raise ValueError(
                    "Upsert insert fields on {} missing conflict field {}".format(
                        schema,
                        field_name,
                    )
                )

            # conflict fields should not be in the update fields (this is more
            # for safety, they should use .update if they want to change them)
            if field_name in update_fields:
                raise ValueError(
                    "Upsert update fields on {} contains conflict field {}".format(
                        schema,
                        field_name,
                    )
                )

        insert_sql, insert_args = self.render_insert_sql(
            schema,
            insert_fields,
            ignore_return_clause=True,
            **kwargs,
        )

        update_sql, update_args = self.render_update_sql(
            schema,
            update_fields,
            query=None,
            only_set_clause=True,
            **kwargs,
        )

        query_str = '{} ON CONFLICT({}) DO UPDATE {}'.format(
            insert_sql,
            ', '.join(map(self.render_field_name_sql, conflict_field_names)),
            update_sql,
        )

        returning_field_names = schema.pk_names
        if returning_field_names:
            # https://www.sqlite.org/lang_returning.html
            query_str += ' RETURNING {}'.format(', '.join(
                map(self.render_field_name_sql, returning_field_names))
            )
            query_args = insert_args + update_args

        r = self._raw(query_str, *query_args, **kwargs)
        if r and returning_field_names:
            if len(returning_field_names) > 1:
                r = r[0]
            else:
                r = r[0][returning_field_names[0]]
        return r

    def _delete(self, schema, query, **kwargs):
        where_query_str, query_args = self.render_sql(schema, query, only_where_clause=True)
        query_str = []
        query_str.append('DELETE FROM')
        query_str.append('  {}'.format(self.render_table_name_sql(schema)))
        query_str.append(where_query_str)
        query_str = "\n".join(query_str)
        ret = self._raw(query_str, *query_args, count_result=True, **kwargs)
        return ret

    def _raw(self, query_str, *query_args, **kwargs):
        """
        **kwargs -- dict
            ignore_result -- boolean -- true to not attempt to fetch results
            fetchone -- boolean -- true to only fetch one result
            count_result -- boolean -- true to return the int count of rows affected
        """
        ret = True
        # http://stackoverflow.com/questions/6739355/dictcursor-doesnt-seem-to-work-under-psycopg2
        #connection = kwargs.get('connection', None)
        with self.connection(**kwargs) as connection:
            cur = connection.cursor()

            ignore_result = kwargs.get('ignore_result', False)
            count_result = kwargs.get('count_result', False)
            one_result = kwargs.get('fetchone', kwargs.get('one_result', False))
            cursor_result = kwargs.get('cursor_result', False)

            if cursor_result:
                # https://stackoverflow.com/a/125140
                cur.arraysize = kwargs.get("arraysize", 500)

            if query_args:
                self.log_for(
                    debug=(["0x{} - {}\n{}", connection, query_str, query_args],),
                    info=([f"{connection} - {query_str}"],)
                )

                try:
                    cur.execute(query_str, query_args)

                except Exception as e:
                    self.raise_error(
                        e,
                        error_args=[
                            query_str,
                            query_args,
                            self.PLACEHOLDER,
                        ]
                    )

            else:
                self.log_info(f"{connection} - {query_str}")
                cur.execute(query_str)

            if cursor_result:
                ret = cur

            elif ignore_result:
                cur.close()

            else:
                if one_result:
                    # https://www.psycopg.org/docs/cursor.html#cursor.fetchone
                    ret = cur.fetchone()

                elif count_result:
                    # https://www.psycopg.org/docs/cursor.html#cursor.rowcount
                    ret = cur.rowcount

                else:
                    # https://www.psycopg.org/docs/cursor.html#cursor.fetchall
                    ret = cur.fetchall()

                cur.close()

            return ret

    def _get(self, schema, query, **kwargs):
        """
        https://www.sqlite.org/lang_select.html
        """
        query_str, query_args = self.render_sql(schema, query)
        return self._raw(query_str, *query_args, **kwargs)

    def _count(self, schema, query, **kwargs):
        query_str, query_args = self.render_sql(schema, query, count_query=True)
        ret = self._raw(query_str, *query_args, **kwargs)
        if ret:
            ret = int(ret[0]['ct'])
        else:
            ret = 0

        return ret

    def _handle_field_error(self, schema, e, **kwargs):
        """
        this will add fields that don't exist in the table if they can be set to NULL,
        the reason they have to be NULL is adding fields to Postgres that can be NULL
        is really light, but if they have a default value, then it can be costly
        """
        current_fields = self._get_fields(schema, **kwargs)
        for field_name, field in schema.fields.items():
            if field_name not in current_fields:
                if field.required:
                    self.log_error(f"Cannot safely add {field_name} on the fly because it is required")
                    return False

                else:
                    query_str = []
                    query_str.append('ALTER TABLE')
                    query_str.append('  {}'.format(self.render_table_name_sql(schema)))
                    query_str.append('ADD COLUMN')
                    query_str.append('  {}'.format(self.render_datatype_sql(field_name, field)))
                    query_str = "\n".join(query_str)
                    self._raw(query_str, ignore_result=True, **kwargs)

        return True

    def _handle_table_error(self, schema, e, **kwargs):
        """
        You can run into a problem when you are trying to set a table and it has a 
        foreign key to a table that doesn't exist, so this method will go through 
        all fk refs and make sure all the tables exist
        """
        if query := kwargs.pop("query", None):
            if schemas := query.schemas:
                for s in schemas:
                    self.log_warning(f"Verifying {schema} query foreign key table: {s}")
                    if not self._handle_table_error(s, e=e, **kwargs):
                        return False

        for field_name, field_val in schema.fields.items():
            s = field_val.schema
            if s:
                self.log_warning(f"Verifying {schema} foreign key table: {s}")
                if not self._handle_table_error(schema=s, e=e, **kwargs):
                    return False

        # now that we know all fk tables exist, create this table
        # !!! This uses the external .set_table so it will run through all the 
        # indexes also
        self.set_table(schema, **kwargs)
        return True

    def render_table_name_sql(self, schema):
        return self.render_field_name_sql(schema)

    def render_field_name_sql(self, name):
        """normalize a non value name for the query

        https://blog.christosoft.de/2012/10/sqlite-escaping-table-acolumn-names/

        :param name: str, the name that should be prepared to be queried
        :returns: the modified name ready to be added to a query string
        """
        return '"{}"'.format(name)

    def render_select_sql(self, schema, query, **kwargs):
        query_str = []
        query_args = []

        only_where_clause = kwargs.get('only_where_clause', False)
        if not only_where_clause:
            query_str.append('SELECT')
            is_count_query = kwargs.get('count_query', False)

            if query.compounds:
                select_fields_str = "*"

            else:
                select_fields = query.fields_select
                if select_fields:
                    distinct_fields = select_fields.options.get(
                        "distinct",
                        select_fields.options.get("unique", False)
                    )
                    distinct = "DISTINCT " if distinct_fields else ""
                    select_fields_str = distinct + ", ".join(
                        (self.render_field_name_sql(f.name) for f in select_fields)
                    )

                else:
                    if is_count_query or select_fields.options.get("all", False):
                        select_fields_str = "*"

                    else:
                        select_fields_str = ", ".join(
                            (self.render_field_name_sql(fname) for fname in schema.fields.keys())
                        )

            if is_count_query:
                query_str.append('  count({}) as ct'.format(select_fields_str))

            else:
                query_str.append('  {}'.format(select_fields_str))

            query_str.append('FROM')

            if not query.compounds:
                query_str.append("  {}".format(self.render_table_name_sql(schema)))

        return query_str, query_args

    def render_compound_sql(self, schema, query, **kwargs):
        query_str = []
        query_args = []

        if query.compounds:
            query_str.append("(")
            for operator, queries in query.compounds:
                for i, query in enumerate(queries):
                    if i > 0:
                        query_str.append(operator.upper())

                    subquery_str, subquery_args = self.render_sql(
                        query.schema,
                        query,
                    )

                    if query.bounds:
                        query_str.append("  (")
                        query_str.append(subquery_str)
                        query_str.append("  )")

                    else:
                        query_str.append(subquery_str)

                    query_args.extend(subquery_args)

            query_str.append(") I")
        return query_str, query_args

    def render_field_raw_sql(self, schema, field):
        return field.name, field.value

    def render_field_sql(self, schema, field):
        format_str = ''
        format_args = []
        is_list = field.is_list
        field_name = field.name
        field_val = field.value

        symbol_map = {
            'in': {'symbol': 'IN', 'list': True},
            'nin': {'symbol': 'NOT IN', 'list': True},
            'eq': {'symbol': '=', 'none_symbol': 'IS'},
            'ne': {'symbol': '!=', 'none_symbol': 'IS NOT'},
            'gt': {'symbol': '>'},
            'gte': {'symbol': '>='},
            'lt': {'symbol': '<'},
            'lte': {'symbol': '<='},
            # https://www.tutorialspoint.com/postgresql/postgresql_like_clause.htm
            # https://www.tutorialspoint.com/sqlite/sqlite_like_clause.htm
            'like': {'symbol': 'LIKE'},
            'nlike': {'symbol': 'NOT LIKE'},
        }
        sd = symbol_map[field.operator]
        symbol = sd['symbol']

        field_kwargs = field.kwargs
        if field_kwargs:
            # kwargs take precedence because None is a perfectly valid field_val
            f = schema.fields[field_name]
            if issubclass(f.type, (datetime.datetime, datetime.date)):
                format_strs = self.render_date_field_sql(
                    field_name,
                    field_kwargs,
                    symbol
                )
                for fname, fvstr, farg in format_strs:
                    if format_str:
                        format_str += ' AND '

                    if is_list:
                        # you can pass in things like day=..., month=... to
                        # date fields, this converts those values to lists to
                        # make sure we can handle something like in_foo(day=1)
                        # and in_foo(day=[1, 2, 3]) the same way
                        farg = make_list(farg)

                        format_str += '{} {} ({})'.format(
                            fname,
                            symbol,
                            ', '.join([fvstr] * len(farg))
                        )
                        format_args.extend(farg)

                    else:
                        format_str += '{} {} {}'.format(fname, symbol, fvstr)
                        format_args.append(farg)

            else:
                raise ValueError(
                    'Field {} does not support extended kwarg values'.format(
                        field_name
                    )
                )

        else:
            if is_list and not isinstance(field_val, Query):
                field_val = make_list(field_val) if field_val else []
                field_name = self.render_field_name_sql(field_name)
                format_val_str = self.PLACEHOLDER

                if field_val:
                    format_str = '{} {} ({})'.format(
                        field_name,
                        symbol,
                        ', '.join([format_val_str] * len(field_val))
                    )
                    format_args.extend(field_val)

                else:
                    # field value is empty, so we need to customize the SQL to
                    # compensate for the empty set since SQL doesn't like empty
                    # sets
                    #
                    # the idea here is this is a condition that will
                    # automatically cause the query to fail but not necessarily 
                    # be an error, the best example is the IN (...) queries, if
                    # you do self.in_foo([]).get() that will fail because the list
                    # was empty, but a value error shouldn't be raised because a
                    # common case is: self.if_foo(Bar.query.is_che(True).pks).get()
                    # which should result in an empty set if there are no rows
                    # where che = TRUE
                    #
                    # https://stackoverflow.com/a/58078468/5006
                    if symbol == "IN":
                        format_str = '{} <> {}'.format(field_name, field_name)

                    elif symbol == "NOT IN":
                        format_str = '{} = {}'.format(field_name, field_name)

                    else:
                        raise ValueError("Unsure what to do here")

            else:
                field_name = self.render_field_name_sql(field_name)
                format_val_str = self.PLACEHOLDER

                # special handling for NULL
                if field_val is None:
                    symbol = sd['none_symbol']

                if isinstance(field_val, Query):
                    subquery_schema = field_val.schema
                    if not subquery_schema:
                        raise ValueError(f"{field_name} subquery has no schema")

                    subquery_sql, subquery_args = self.render_sql(
                        subquery_schema,
                        field_val
                    )

                    format_str = '{} {} ({})'.format(
                        field_name,
                        symbol,
                        subquery_sql
                    )
                    format_args.extend(subquery_args)

                else:
                    format_str = '{} {} {}'.format(
                        field_name,
                        symbol,
                        format_val_str,
                    )
                    format_args.append(field_val)

        return format_str, format_args

    def render_where_sql(self, schema, query, **kwargs):
        query_str = []
        query_args = []

        if query.fields_where:
            query_str.append('WHERE')
            or_clause = False

            for i, field in enumerate(query.fields_where):
                if i > 0:
                    query_str.append('OR' if or_clause else 'AND')

                field_str = ''
                field_args = []

                if field.raw:
                    field_str, field_args = self.render_field_raw_sql(
                        schema,
                        field,
                    )

                else:
                    field_str, field_args = self.render_field_sql(
                        schema,
                        field,
                    )

                if field.or_clause:
                    if not or_clause:
                        query_str.append("(")
                        or_clause = True

                query_str.append('  {}'.format(field_str))
                query_args.extend(field_args)

                if or_clause:
                    if not field.or_clause:
                        query_str.append(")")
                        or_clause = False

        return query_str, query_args

    def render_sort_sql(self, schema, query, **kwargs):
        query_str = []
        query_args = []

        if query.fields_sort:
            query_str.append('ORDER BY')

            query_sort_str = []
            for field in query.fields_sort:
                sort_dir_str = 'ASC' if field.direction > 0 else 'DESC'
                if field.value:
                    field_sort_str, field_sort_args = self.render_sort_field_sql(
                        field.name,
                        field.value,
                        sort_dir_str
                    )
                    query_sort_str.append(field_sort_str)
                    query_args.extend(field_sort_args)

                else:
                    query_sort_str.append('  {} {}'.format(field.name, sort_dir_str))

            query_str.append(',\n'.join(query_sort_str))

        return query_str, query_args

    def render_bounds_sql(self, schema, query, **kwargs):
        """
        https://www.postgresql.org/docs/current/queries-limit.html
        https://www.sqlite.org/lang_select.html#the_limit_clause
        """
        query_str = []
        query_args = []

        bounds = query.bounds
        fetchone = kwargs.get("fetchone", False)
        if bounds or fetchone:
            if fetchone:
                limit = 1
                offset = bounds.offset

            else:
                if bounds.has_limit():
                    limit, offset = bounds.get()

                else:
                    limit = self.LIMIT_NONE
                    offset = bounds.offset

            query_str.append(f'LIMIT {limit} OFFSET {offset}')

        return query_str, query_args

    def render_sql(self, schema, query, **kwargs):
        """
        convert the query instance into SQL

        this is the glue method that translates the generic Query() instance to
        the SQL specific query, this is where the magic happens

        https://www.sqlite.org/lang_select.html

        :param **kwargs:
            - count_query: bool, True if this is a count query SELECT
            - only_where_clause, bool, True to only return after WHERE ...
        :returns: tuple[str, list[Any]], (query_str, query_args)
        """
        query_str = []
        query_args = []

        select_str, select_args = self.render_select_sql(
            schema,
            query,
            **kwargs
        )
        query_str.extend(select_str)
        query_args.extend(select_args)

        compound_str, compound_args = self.render_compound_sql(
            schema,
            query,
            **kwargs
        )
        query_str.extend(compound_str)
        query_args.extend(compound_args)

        where_str, where_args = self.render_where_sql(
            schema,
            query,
            **kwargs
        )
        query_str.extend(where_str)
        query_args.extend(where_args)

        sort_str, sort_args = self.render_sort_sql(
            schema,
            query,
            **kwargs
        )
        query_str.extend(sort_str)
        query_args.extend(sort_args)

        limit_str, limit_args = self.render_bounds_sql(
            schema,
            query,
            **kwargs
        )
        query_str.extend(limit_str)
        query_args.extend(limit_args)

        query_str = "\n".join(query_str)
        return query_str, query_args

    def render_insert_sql(self, schema, fields, **kwargs):
        """
        https://www.sqlite.org/lang_insert.html
        """
        field_names = []
        query_vals = []
        for field_name, field_val in fields.items():
            field_names.append(field_name)
            query_vals.append(field_val)

        query_str = self.render_inserts_sql(schema, field_names, **kwargs)

        if not kwargs.get("ignore_return_clause", False):
            # https://www.sqlite.org/lang_returning.html
            pk_name = schema.pk_name
            if pk_name:
                query_str += ' RETURNING {}'.format(self.render_field_name_sql(pk_name))

        return query_str, query_vals

    def render_inserts_sql(self, schema, field_names, **kwargs):
        """
        https://www.sqlite.org/lang_insert.html
        """
        sql_names = []
        sql_formats = []
        for field_name in field_names:
            sql_names.append(self.render_field_name_sql(field_name))
            sql_formats.append(self.PLACEHOLDER)

        return "INSERT INTO {} ({}) VALUES ({})".format(
            self.render_table_name_sql(schema),
            ", ".join(sql_names),
            ", ".join(sql_formats),
        )

    def render_update_sql(self, schema, fields, query, **kwargs):
        query_str = ''
        query_args = []

        if not kwargs.get("only_set_clause", False):
            query_str = 'UPDATE {} '.format(self.render_table_name_sql(schema))

        field_str = []
        for field_name, field_val in fields.items():
            field_str.append('{} = {}'.format(self.render_field_name_sql(field_name), self.PLACEHOLDER))
            query_args.append(field_val)

        query_str += 'SET {}'.format(',\n'.join(field_str))

        if query:
            where_query_str, where_query_args = self.render_sql(schema, query, only_where_clause=True)
            query_str += ' {}'.format(where_query_str)
            query_args.extend(where_query_args)

        return query_str, query_args

    def render_datatype_sql(self, field_name, field):
        """Returns the SQL for a given field with full type information

        http://www.sqlite.org/datatype3.html
        https://www.postgresql.org/docs/current/datatype.html

        :param field_name: str, the field's name
        :param field: Field instance, the configuration for the field
        :returns: str, the complete field datatype SQL (eg, foo BOOL NOT NULL)
        """
        field_type = ""
        interface_type = field.interface_type

        if issubclass(interface_type, bool):
            field_type = self.render_datatype_bool_sql(field_name, field)

        elif issubclass(interface_type, int):
            field_type = self.render_datatype_int_sql(field_name, field)

        elif issubclass(interface_type, str):
            field_type = self.render_datatype_str_sql(field_name, field)

        elif issubclass(interface_type, datetime.datetime):
            field_type = self.render_datatype_datetime_sql(field_name, field)

        elif issubclass(interface_type, datetime.date):
            field_type = self.render_datatype_date_sql(field_name, field)

        elif issubclass(interface_type, dict):
            field_type = self.render_datatype_dict_sql(field_name, field)

        elif issubclass(interface_type, (float, decimal.Decimal)):
            field_type = self.render_datatype_float_sql(field_name, field)

        elif issubclass(interface_type, (bytearray, bytes)):
            field_type = self.render_datatype_bytes_sql(field_name, field)

        elif issubclass(interface_type, uuid.UUID):
            field_type = self.render_datatype_uuid_sql(field_name, field)

        else:
            raise ValueError('Unknown python type: {} for field: {}'.format(
                interface_type.__name__,
                field_name,
            ))

        field_type += ' ' + self.render_datatype_required_sql(field_name, field)

        if field.is_pk():
            field_type += ' PRIMARY KEY'

        else:
            if field.is_ref():
                field_type += ' ' + self.render_datatype_ref_sql(field_name, field)

        return '{} {}'.format(self.render_field_name_sql(field_name), field_type)

    def render_datatype_bool_sql(self, field_name, field, **kwargs):
        return 'BOOL'

    def render_datatype_int_sql(self, field_name, field, **kwargs):
        return 'INTEGER'

    def render_datatype_str_sql(self, field_name, field, **kwargs):
        fo = field.interface_options
        field_type = kwargs.get("datatype", 'TEXT')
        size_info = field.size_info()

        # https://www.sqlitetutorial.net/sqlite-check-constraint/
        if 'size' in size_info["original"]:
            field_type += f" CHECK(length({field_name}) = {size_info['size']})"

        elif 'max_size' in size_info["original"]:
            if "min_size" in size_info["original"]:
                field_type += f" CHECK(length({field_name}) >= {size_info['original']['min_size']}"
                field_type += " AND "
                field_type += f"length({field_name}) <= {size_info['original']['max_size']})"

            else:
                field_type += f" CHECK(length({field_name}) <= {size_info['size']})"

        return field_type

    def render_datatype_date_sql(self, field_name, field):
        return 'DATE'

    def render_datatype_float_sql(self, field_name, field, **kwargs):
        return 'REAL'

    def render_datatype_bytes_sql(self, field_name, field, **kwargs):
        return 'BLOB'

    def render_datatype_required_sql(self, field_name, field, **kwargs):
        return 'NOT NULL' if field.required else 'NULL'

    def render_datatype_ref_sql(self, field_name, field, **kwargs):
        ref_s = field.schema
        if field.required: # strong ref, it deletes on fk row removal
            format_str = 'REFERENCES {} ({}) ON UPDATE CASCADE ON DELETE CASCADE'

        else: # weak ref, it sets column to null on fk row removal
            format_str = 'REFERENCES {} ({}) ON UPDATE CASCADE ON DELETE SET NULL'

        ret = format_str.format(
            self.render_table_name_sql(ref_s),
            self.render_field_name_sql(ref_s.pk.name)
        )

        return ret

    def render(self, schema, query, **kwargs):
        """Render the query

        :param schema: Schema, the query schema
        :param query: Query, the query to render
        :param **kwargs: named arguments
            placeholders: boolean, True if place holders should remain
        :returns: string if placeholders is False, (string, list) if placeholders is True
        """
        sql, sql_args = self.render_sql(schema, query)
        placeholders = kwargs.get("placeholders", kwargs.get("placeholder", False))

        if not placeholders:
            for sql_arg in sql_args:
                if sql_arg is None:
                    sa = 'NULL'

                else:
                    sa = String(sql_arg)
                    if not sa.isnumeric():
                        sa = "'{}'".format(sa)

                sql = sql.replace(self.PLACEHOLDER, sa, 1)

        return (sql, sql_args) if placeholders else sql




# -*- coding: utf-8 -*-
from __future__ import unicode_literals, division, print_function, absolute_import


class Error(Exception):
    """all prom errors will inherit from this base class

    :param e: Exception, the exception instance that is getting wrapped
    :param **kwargs:
        - message: by default it will use str(e) but you can customize it by
          passing in "message", you can see in the source code that message is
          usually the first argument, it doesn't have a name, it's just args[0],
          https://github.com/python/cpython/blob/3.11/Objects/exceptions.c
    """
    def __init__(self, e, **kwargs):
        self.e = e
        super(Error, self).__init__(kwargs.get("message", str(e)))

    def unwrapped_e(self):
        """Find the first unwrapped error (ie, unwind to the original error)"""
        e = self
        # unwind to the original error
        while isinstance(e, InterfaceError):
            e = e.e
        return e


class InterfaceError(Error):
    """specifically for wrapping SQLite and Postgres errors

    see Interface.create_error() for how InterfaceError instances are created
    """
    def __str__(self):
        """Postgres returns multi-line errors, this switches to just return the
        first line since the other lines aren't usually helpful, if you want the
        full message just call .unwrapped_e()"""
        s = super().__str__()
        lines = s.splitlines()
        return lines[0] if lines else ""


class TableError(InterfaceError):
    pass


class FieldError(InterfaceError):
    pass


class UniqueError(FieldError):
    pass


class CloseError(InterfaceError):
    pass


class PlaceholderError(InterfaceError):
    """This is raised when there is a raw query mismatch between the placeholders
    and the passed in arguments, see Interface._raw

    https://github.com/Jaymon/prom/issues/74
    """
    def __init__(self, e, query_str, query_args, placeholder, **kwargs):
        if "message" not in kwargs:
            count = query_str.count(placeholder)
            kwargs["message"] = "Query has {} placeholder(s) but {} value(s)".format(
                count,
                len(query_args),
            )

        super().__init__(e, **kwargs)

        self.query_str = query_str
        self.query_args = query_args
        self.placeholder = placeholder




# -*- coding: utf-8 -*-
from __future__ import unicode_literals, division, print_function, absolute_import
import importlib
import heapq
import itertools
import os
import sys
import codecs
from contextlib import contextmanager

from .compat import *


class Stream(object):
    """In the CLI we either want to print to stdout or write to a file, this wrapper
    does that, if path is empty then anything you write to it will write to stdout
    """
    def __init__(self, path=""):
        self.path = path
        if path:
            self.path = os.path.abspath(os.path.expanduser(str(path)))
            self.stream = codecs.open(self.path, encoding='utf-8', mode='w+')
        else:
            self.stream = sys.stdout

    def __getattr__(self, k):
        return getattr(self.stream, k)

    @contextmanager
    def open(self):
        yield self
        if self.path:
            self.stream.close()

    def write_line(self, line, count=1):
        """writes the line and count newlines after the line"""
        self.write(line)
        self.write_newlines(count)

    def write_newlines(self, count=1):
        """writes count newlines"""
        for c in range(count):
            self.write("\n")


def get_objects(classpath, calling_classpath=""):
    """
    given a classpath like foo.bar.Baz return module foo.bar and class Baz
    objects

    .. seealso::
        https://docs.python.org/2.5/whatsnew/pep-328.html
        https://www.python.org/dev/peps/pep-0328/

    :param classpath: string, the full python class path (includes modules), a classpath
        is something like foo.bar.Che where Che is the class definied in the foo.bar
        module
    :param calling_classpath: string, if classpath is relative (eg, ..foo.Bar) then
        this is needed to resolve the relative classpath, it is usually the path of
        the module that is calling get_objects()
    :returns: tuple, (module, class)
    """
#     if classpath.startswith("."):
#         rel_count = len(re.match("^\.+", classpath).group(0))
#         if calling_classpath:
#             calling_count = calling_classpath.count(".")
#             if rel_count > calling_count:
#                 raise ValueError(
#                     "Attempting relative import passed calling_classpath {}".format(
#                         calling_classpath
#                     )
#                 )
# 
#             bits = calling_classpath.rsplit('.', rel_count)
#             parent_classpath = bits[0]
#             classpath = ".".join([parent_classpath, classpath[rel_count:]])
# 
#         else:
#             raise ValueError("Attempting relative import without calling_classpath")
# 
    module_name, class_name = classpath.rsplit('.', 1)
    module = importlib.import_module(module_name, calling_classpath)
    try:
        klass = getattr(module, class_name)
    except AttributeError:
        raise AttributeError("module {} has no attribute {} parsing {}".format(
            module.__name__,
            class_name,
            classpath
        ))

    return module, klass


def make_list(vals):
    """make vals a list, no matter what

    :param vals: list, vals should always be a list or tuple (eg, *args passed as args)
    """
    ret = []
    if isinstance(vals, basestring):
        ret.append(vals)

    else:
        try:
            for val in vals:
                if isinstance(val, basestring):
                    ret.append(val)

                elif isinstance(val, (list, tuple)):
                    ret.extend(val)

                else:
                    try:
                        r = list(val)
                        ret.extend(r)

                    except TypeError:
                        ret.append(val)

        except TypeError:
            # TypeError: * is not iterable
            ret.append(vals)

    return ret


def make_dict(fields, fields_kwargs):
    """lot's of methods take a dict or kwargs, this combines those

    Basically, we do a lot of def method(fields, **kwargs) and we want to merge
    those into one super dict with kwargs taking precedence, this does that

    fields -- dict -- a passed in dict
    fields_kwargs -- dict -- usually a **kwargs dict from another function

    return -- dict -- a merged fields and fields_kwargs
    """
    ret = {}
    if fields:
        ret.update(fields)

    if fields_kwargs:
        ret.update(fields_kwargs)

    return ret


def make_hash(*mixed):
    s = ""
    for m in mixed:
        s += String(m)
    # http://stackoverflow.com/questions/5297448/how-to-get-md5-sum-of-a-string
    return String(s).md5()




# -*- coding: utf-8 -*-
from __future__ import unicode_literals, division, print_function, absolute_import
import sys
import os
import datetime
import decimal
import logging
from contextlib import contextmanager
import uuid
import weakref

from datatypes import (
    LogMixin,
    Stack,
)

# first party
from ..query import Query
from ..exception import (
    InterfaceError,
    UniqueError,
    TableError,
    FieldError,
    UniqueError,
    CloseError,
)

from ..compat import *
from ..utils import make_list


logger = logging.getLogger(__name__)


class ConnectionABC(LogMixin):
    """Subclasses should extend Connection and implement the methods in this class"""
    def _transaction_start(self):
        """Called when the first transaction is started"""
        pass

    def _transaction_starting(self, tx):
        """Called when a nested transaction is started"""
        pass

    def _transaction_ignoring(self, tx):
        """Called when nested transaction is ignored instead of started"""
        pass

    def _transaction_stop(self):
        """Called when the last transaction is stopped"""
        pass

    def _transaction_stopping(self, tx):
        """Called when a nested transaction is stopped"""
        pass

    def _transaction_fail(self):
        """Called when the last transaction is failed"""
        pass

    def _transaction_failing(self, tx):
        """Called when a nested transaction is failed"""
        pass


class Connection(ConnectionABC):
    """holds common methods that all raw connections should have"""
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # counting semaphore, greater than 0 if in a transaction, 0 if no current
        # transaction.
        #
        # This will be incremented everytime transaction_start() is called, and
        # decremented everytime transaction_stop() is called.
        #
        # transaction_fail will set this back to 0 and rollback the transaction
        #
        # Holds the active transactions
        self.transactions = Stack()

    def transaction_count(self):
        """How many active transactions there currently are"""
        return len(self.transactions)

    def transaction_names(self):
        return " > ".join((r["name"] for r in reversed(self.transactions)))

    def transaction_name(self, **kwargs):
        """generate a random transaction name for use in start_transaction() and
        fail_transaction()

        :param prefix: str, to better track transactions in logs you can give a
            prefix name that will be prepended to the auto-generated name
        """
        if not (name := kwargs.get("name", "")):
            suffix = str(uuid.uuid4())[-5:]
            prefix = kwargs.get("prefix", "") or "p"
            name = f"{prefix}_{suffix}"
        return name

    def transaction_create(self, **kwargs):
        """Create a new transaction dict that will be placed on the .transactions
        stack

        :param **kwargs:
            - nest: bool, True if this (and children unless passed in) will run
                nested transactions. If this is False then subsequent calls to
                .transaction_start will be ignored unless nest=True is passed in
                again
            - name: str, the transaction name
            - prefix: str, used to create a transaction name (see
              .transaction_name)
        :returns: dict[str], the created transaction with keys:
            - nest: bool, the value of kwargs["nest"] or of
              .transaction_current()["nest"]
            - name: str, the tx name
            - ignored: bool, True if this tx is going to be ignored
            - index: int, the depth of the transaction
        """
        name = self.transaction_name(**kwargs)
        current_tx = self.transaction_current()
        nest = kwargs.get("nest", current_tx.get("nest", True))

        ignored = False
        if current_tx:
            ignored = not nest

        else:
            ignored = False

        index = self.transaction_count() + 1

        return {
            "nest": nest,
            "name": name,
            "ignored": ignored,
            "index": index
        }

    def transaction_current(self):
        """Returns the current transaction dict, or empty dict if no tx"""
        return self.transactions[-1] if self.transactions else {}

    def transaction_exists(self):
        """return true if currently in a transaction

        this was previously named .in_transaction but it turns out SQLite has a
        property with that name
        """
        return self.transaction_count() > 0

    def transaction_start(self, **kwargs):
        """start a transaction

        this will increment transaction semaphore and pass it to _transaction_start()
        """
        tx = self.transaction_create(**kwargs)
        self.transactions.push(tx)

        transaction_count = tx["index"]
        if tx["ignored"]:
            self.log_debug([
                f"{transaction_count}.",
                f"Ignoring {self} transaction {self.transaction_names()}",
            ])
            self._transaction_ignoring(tx)

        else:
            self.log_debug([
                f"{transaction_count}.", 
                f"Start {self} transaction {self.transaction_names()}",
            ])
            if transaction_count == 1:
                self._transaction_start()

            else:
                self._transaction_starting(tx)

        return transaction_count

    def transaction_stop(self):
        """stop/commit a transaction if ready"""
        transaction_count = self.transaction_count()
        if transaction_count > 0:
            tx = self.transactions.pop()
            if not tx["ignored"]:
                self.log_debug([
                    f"{transaction_count}.", 
                    f"Stopping {self} transaction {self.transaction_names()}",
                ])

                if transaction_count == 1:
                    self._transaction_stop()

                else:
                    self._transaction_stopping(tx)

        return self.transaction_count

    def transaction_fail(self):
        """rollback a transaction if currently in one"""
        transaction_count = self.transaction_count()
        if transaction_count > 0:
            tx = self.transactions.pop()
            if not tx["ignored"]:
                self.log_debug([
                    f"{transaction_count}.",
                    f"Failing {self} transaction {self.transaction_names()}",
                ])

                if transaction_count == 1:
                    self._transaction_fail()

                else:
                    self._transaction_failing(tx)

    def __str__(self):
        return f"0x{id(self):02x}"


class InterfaceABC(LogMixin):
    """This is just a convenience abstract base class so child interfaces can easily
    see what methods they need to implement. They should extend Interface and then
    implement the methods in this class"""
    def _connect(self, config):
        raise NotImplementedError()

    def _configure_connection(self, **kwargs):
        """This is ran immediately after a successful .connect()"""
        pass

    def _free_connection(self, connection):
        pass

    def _get_connection(self):
        raise NotImplementedError()

    def _close(self):
        raise NotImplementedError()

    def _readonly(self, readonly, **kwargs):
        raise NotImplementedError()

    def _raw(self, query_str, *query_args, **kwargs):
        raise NotImplementedError()

    def _set_table(self, schema, **kwargs):
        raise NotImplementedError()

    def _get_tables(self, table_name, **kwargs):
        raise NotImplementedError()

    def _delete_table(self, schema):
        raise NotImplementedError()

    def _get_fields(self, table_name, **kwargs):
        raise NotImplementedError()

    def _get_indexes(self, schema, **kwargs):
        raise NotImplementedError()

    def _set_index(self, schema, name, field_names, **kwargs):
        raise NotImplementedError()

    def _insert(self, schema, fields, **kwargs):
        raise NotImplementedError()

    def _inserts(self, schema, field_names, field_values, **kwargs):
        raise NotImplementedError()

    def _update(self, schema, fields, query, **kwargs):
        raise NotImplementedError()

    def _upsert(self, schema, insert_fields, update_fields, conflict_field_names, **kwargs):
        raise NotImplementedError()

    def _delete(self, schema, query, **kwargs):
        raise NotImplementedError()

    def _get(self, schema, query, **kwargs):
        raise NotImplementedError()

    def _count(self, schema, query, **kwargs):
        raise NotImplementedError()

    def render(self, schema, query, **kwargs):
        """Render the query in a way that the interface can interpret it

        so in a SQL interface, this would render SQL, this is mainly used for
        debugging

        :param query: Query, the Query instance to render
        :param **kwargs: any named arguments
        :returns: mixed
        """
        raise NotImplementedError()

    def _handle_unique_error(self, e, **kwargs):
        return False

    def _handle_field_error(self, e, **kwargs):
        return False

    def _handle_table_error(self, e, **kwargs):
        return False

    def _handle_general_error(self, e, **kwargs):
        return False

    def _handle_close_error(self, e, **kwargs):
        self.close()
        return True


class Interface(InterfaceABC):

    connected = False
    """true if a connection has been established, false otherwise"""

    config = None
    """a config.Connection() instance"""

    InterfaceError = InterfaceError
    UniqueError = UniqueError
    TableError = TableError
    FieldError = FieldError
    UniqueError = UniqueError
    CloseError = CloseError

    @classmethod
    def configure(cls, config):
        host = config.host
        if host:
            db = config.database
            config.database = db.strip("/")
        return config

    def __init__(self, config=None):
        self.config = config

        # enables cleanup of open sockets even if the object isn't correctly garbage collected
        weakref.finalize(self, self.__del__)

    def __del__(self):
        """Whenever this gets garbage collected close everything. This is also the
        method for weakref.finalize"""
        self.close()

    def connect(self, config=None, *args, **kwargs):
        """
        connect to the interface

        this will set the raw db connection to self.connection

        *args -- anything you want that will help the db connect
        **kwargs -- anything you want that the backend db connection will need to actually connect
        """
        if self.connected:
            return self.connected

        if config:
            self.config = config

        try:
            self._connect(self.config)
            self.connected = True

        except Exception as e:
            self.connected = False
            self.raise_error(e)

        self.log_debug(f"Connected {self.config.interface_name}")
        self.configure_connection()
        return self.connected

    def reconnect(self):
        self.close()
        self.connect()

    def configure_connection(self, **kwargs):
        kwargs.setdefault("prefix", "configure_connection")
        self.execute(
            self._configure_connection,
            **kwargs
        )

    def get_connection(self):
        """Any time you need a connection it should be retrieved through .connection,
        and that method uses this method

        :returns: Connection instance
        """
        if not self.is_connected():
            self.connect()

        connection = self._get_connection()

        if connection.closed:
            # we've gotten into a bad state so let's try reconnecting
            self.reconnect()
            connection = self._get_connection()

        connection.interface = self

        self.log_debug(
            f"Getting {self.config.interface_name} connection {connection}"
        )
        return connection

    def free_connection(self, connection):
        """When .connection is done with a connection it calls this method"""
        #connection.interface = None
        if self.is_connected():
            self.log_debug(
                f"Freeing {self.config.interface_name} connection {connection}"
            )
            self._free_connection(connection)

    def is_connected(self):
        return self.connected

    def close(self):
        """close an open connection"""
        if not self.connected: return True

        self._close()
        self.connected = False
        self.log_debug(f"Closed Connection {self.config.interface_name}")
        return True

    def readonly(self, readonly=True, **kwargs):
        """Make the connection read only (pass in True) or read/write (pass in False)

        :param readonly: boolean, True if this connection should be readonly, False
            if the connection should be read/write
        """
        self.log_warning([
            f"Setting interface {self.config.interface_name}",
            f"to readonly={readonly}",
        ])
        self.config.readonly = readonly

        if self.connected:
            kwargs.setdefault("prefix", "readonly")
            self.execute(
                self._readonly,
                readonly,
                **kwargs
            )
            #self._readonly(readonly, **kwargs)

    @contextmanager
    def connection(self, connection=None, **kwargs):
        """Any time you need a connection you should use this context manager, this
        is the only place that wraps exceptions in InterfaceError, so all connections
        should go through this method or .transaction if you need to start a transaction

        :Example:
            with self.connection(**kwargs) as connection:
                # do something with connection
        """
        free_connection = False
        try:
            if connection:
                if connection.closed:
                    self.log_warning(
                        "Passed in connection is closed and must be refreshed"
                    )
                    if connection.transaction_exists():
                        self.log_warning("Closed connection had open transactions!")

                    connection = None

                else:
                    self.log_debug(
                        f"Connection call using existing connection {connection}"
                    )
                    yield connection

            if connection is None:
                free_connection = True
                connection = self.get_connection()
                yield connection

        except Exception as e:
            self.raise_error(e)

        finally:
            if free_connection and connection:
                self.free_connection(connection)

            else:
                self.log_debug(
                    f"Connection call NOT freeing existing connection {connection}"
                )

    @contextmanager
    def transaction(self, connection=None, **kwargs):
        """A simple context manager useful for when you want to wrap a bunch of
        db calls in a transaction, this is used internally for any write statements

        :Example:
            with self.transaction() as connection
                # do a bunch of calls
            # those db calls will be committed by this line
        """
        with self.connection(connection) as connection:
            connection.transaction_start(**kwargs)
            try:
                yield connection

            except Exception:
                connection.transaction_fail()
                raise

            else:
                connection.transaction_stop()

    def execute_write(self, callback, *args, **kwargs):
        """Any write statements will use this method

        CREATE, DELETE, DROP, INSERT, or UPDATE (collectively "write statements")
        """
        kwargs.setdefault("execute_in_transaction", True)

        return self.execute(callback, *args, **kwargs)

    def execute_read(self, callback, *args, **kwargs):
        """Any write statements will use this method

        SELECT (collectively "read statements")
        """
        with self.connection(**kwargs) as connection:
            kwargs["connection"] = connection

            kwargs.setdefault(
                "execute_in_transaction",
                connection.transaction_exists()
            )

            return self.execute(callback, *args, **kwargs)

    def execute(self, callback, *args, **kwargs):
        """Internal method. Execute the callback with args and kwargs, retrying
        the query if an error is raised that it thinks it successfully handled

        This is called by .execute_write, .execute_read, and .query

        :param callback: callable, this will be run at-most twice
        :param *args: passed directly to callback as *args
        :param **kwargs: passed to callback as **kwargs, can have values added
        :returns: mixed, whatever the callback returns
        """
        prefix = kwargs.pop("prefix", callback.__name__)

        try:
            return self._execute(callback, *args, prefix=prefix, **kwargs)

        except Exception as e:
            if self.handle_error(e=e, prefix=prefix, **kwargs):
                return self._execute(
                    callback,
                    *args,
                    prefix=f"{prefix}_retry",
                    **kwargs
                )

            else:
                self.raise_error(e)

    def _execute(self, callback, *args, **kwargs):
        """Internal method. Called by .execute"""
        in_transaction = kwargs.get("execute_in_transaction", False)

        if in_transaction:
            with self.transaction(**kwargs) as connection:
                kwargs["connection"] = connection
                return callback(*args, **kwargs)

        else:
            with self.connection(**kwargs) as connection:
                kwargs["connection"] = connection
                return callback(*args, **kwargs)

    def has_table(self, table_name, **kwargs):
        """
        check to see if a table is in the db

        table_name -- string -- the table to check
        return -- boolean -- True if the table exists, false otherwise
        """
        kwargs.setdefault("prefix", "has_table")
        tables = self.execute_read(
            self.get_tables,
            table_name,
            **kwargs
        )
        return len(tables) > 0

    def get_tables(self, table_name="", **kwargs):
        """
        get all the tables of the currently connected db

        table_name -- string -- if you would like to filter the tables list to only include matches with this name
        return -- list -- a list of table names
        """
        kwargs.setdefault("prefix", "get_tables")
        return self.execute_read(
            self._get_tables,
            str(table_name),
            **kwargs
        )

    def set_table(self, schema, **kwargs):
        """
        add the table to the db

        :param schema: Schema instance, contains all the information about the table
        """
        kwargs.setdefault("prefix", "set_table")
        with self.transaction(**kwargs) as connection:
            kwargs['connection'] = connection

            self._set_table(schema=schema, **kwargs)

            for index_name, index in schema.indexes.items():
                self._set_index(
                    schema=schema,
                    name=index.name,
                    field_names=index.field_names,
                    connection=connection,
                    **index.options,
                )

    def unsafe_delete_table(self, schema, **kwargs):
        """wrapper around delete_table that matches the *_tables variant and denotes
        that this is a serious operation

        remove a table matching schema from the db

        :param schema: Schema instance, the table to delete
        """
        kwargs.setdefault("prefix", "unsafe_delete_table")
        self.execute_write(
            self._delete_table,
            schema=schema,
            **kwargs
        )
        return True

    def unsafe_delete_tables(self, **kwargs):
        """Removes all the tables from the db

        https://github.com/Jaymon/prom/issues/75
        """
        kwargs.setdefault("prefix", "unsafe_delete_tables")
        with self.transaction(**kwargs) as connection:
            kwargs['connection'] = connection
            kwargs.setdefault('nest', False)
            for table_name in self.get_tables(**kwargs):
                self._delete_table(table_name, **kwargs)
        return True

    def get_indexes(self, schema, **kwargs):
        """
        get all the indexes

        schema -- Schema()

        return -- dict -- the indexes in {indexname: fields} format
        """
        kwargs.setdefault("prefix", "get_indexes")
        return self.execute_read(
            self._get_indexes,
            schema=schema,
            **kwargs
        )

    def set_index(self, schema, name, field_names, **kwargs):
        """
        add an index to the table

        schema -- Schema()
        name -- string -- the name of the index
        field_names -- array -- the fields the index should be on
        **index_options -- dict -- any index options that might be useful to create the index
        """
        kwargs.setdefault("prefix", "set_index")
        self.execute_write(
            self._set_index,
            schema=schema,
            name=name,
            field_names=field_names,
            **kwargs,
        )
        return True

    def insert(self, schema, fields, **kwargs):
        """Persist fields into the db

        :param schema: Schema instance, the table the query will run against
        :param fields: dict, the fields {field_name: field_value} to persist
        :param **kwargs: passed through
        :returns: mixed, will return the primary key values
        """
        kwargs.setdefault("prefix", "insert")
        return self.execute_write(
            self._insert,
            schema=schema,
            fields=fields,
            **kwargs
        )

    def inserts(self, schema, field_names, field_rows, **kwargs):
        """Persist the field names found in all the field rows

        :param schema: Schema
        :param field_names: list, the field names that will be checked for each
            row in field_rows and used to turn each dict in field_rows to a tuple
        :param field_rows: Sequence, if an iterator of dict instances with keys found
            in field_names, if a key is missing it will have None set as the value,
            if it's an iterator of values then the tuple value ordering should match
            with field_names
        :param **kwargs: passed through
        :returns: bool, True if the query executed successfully
        """
        kwargs.setdefault("prefix", "inserts")

        def field_values(field_names, field_rows):
            for fields in field_rows:
                if isinstance(fields, Mapping):
                    row = []
                    for field_name in field_names:
                        row.append(fields.get(field_name, None))
                    yield row

                else:
                    yield fields

        self.execute_write(
            self._inserts,
            schema=schema,
            field_names=field_names,
            field_values=field_values(field_names, field_rows),
            **kwargs,
        )
        return True

    def update(self, schema, fields, query, **kwargs):
        """Persist the query.fields into the db that match query.fields_where

        :param schema: Schema instance, the table the query will run against
        :param fields: dict, the fields {field_name: field_value} to persist
        :param query: Query instance, will be used to create the where clause
        :param **kwargs: passed through
        :returns: int, how many rows where updated
        """
        kwargs.setdefault("prefix", "update")
        return self.execute_write(
            self._update,
            schema=schema,
            fields=fields,
            query=query,
            **kwargs,
        )

    def upsert(self, schema, insert_fields, update_fields, conflict_field_names, **kwargs):
        """Perform an upsert (insert or update) on the table

        :param schema: Schema instance, the table the query will run against
        :param insert_fields: dict, these are the fields that will be inserted
        :param update_fields: dict, on a conflict with the insert_fields, these
            fields will instead be used to update the row
        :param conflict_field_names: list, the field names that will decide if
            an insert or update is performed
        :param **kwargs: anything else
        :returns: mixed, the primary key
        """
        kwargs.setdefault("prefix", "upsert")
        return self.execute_write(
            self._upsert,
            schema=schema,
            insert_fields=insert_fields,
            update_fields=update_fields,
            conflict_field_names=conflict_field_names,
            **kwargs,
        )

    def delete(self, schema, query, **kwargs):
        """delete matching rows according to query filter criteria

        :param schema: Schema instance, the table the query will run against
        :param query: Query instance, the filter criteria, this will fail if empty
        :returns: int, how many rows were deleted ... I think
        """
        if not query or not query.fields_where:
            raise ValueError('aborting delete because there is no where clause')

        kwargs.setdefault("prefix", "delete")
        return self.execute_write(
            self._delete,
            schema=schema,
            query=query,
            **kwargs
        )

    def raw(self, query_str, *query_args, **kwargs):
        """
        run a raw query on the db

        query_str -- string -- the query to run
        *query_args -- if the query_str is a formatting string, pass the values in this
        **kwargs -- any query options can be passed in by using key=val syntax
        """
        kwargs.setdefault("prefix", "raw")
        return self.execute(
            self._raw,
            query_str,
            *query_args,
            **kwargs
        )

    def get_fields(self, table_name, **kwargs):
        kwargs.setdefault("prefix", "get_fields")
        return self.execute_read(
            self._get_fields,
            str(table_name),
            **kwargs
        )

    def get_one(self, schema, query=None, **kwargs):
        """get one row from the db matching filters set in query

        :param schema: Schema instance, the table the query will run against
        :param query: Query instance, the filter criteria
        :return: dict, the matching row
        """
        kwargs.setdefault("prefix", "get_one")
        return self.get(schema, query, fetchone=True, **kwargs) or {}

    def get(self, schema, query=None, **kwargs):
        """get matching rows from the db matching filters set in query

        :param schema: Schema instance, the table the query will run against
        :param query: Query instance, the filter criteria
        :returns: list, a list of matching dicts
        """
        kwargs.setdefault("prefix", "get")
        ret = self.execute_read(
            self._get,
            schema=schema,
            query=query or Query(),
            **kwargs
        )
        return ret or []

    def count(self, schema, query=None, **kwargs):
        """count matching rows according to query filter criteria

        :param schema: Schema instance, the table the query will run against
        :param query: Query instance, the filter criteria
        :returns: list, a list of matching dicts
        """
        kwargs.setdefault("prefix", "count")
        ret = self.execute_read(
            self._count,
            schema=schema,
            query=query or Query(),
            **kwargs
        )
        return int(ret) if ret else 0

    def handle_error(self, e, **kwargs):
        """Try and handle the error, return False if the error can't be handled

        :param e: Exception, the caught exception
        :param **kwargs:
            - schema: Schema, this does not have to be there, but usually is
        :returns: bool, True if the error was handled, False if it wasn't
        """
        ret = False

        prefix = kwargs.get("prefix", "")
        self.log_warning(["Handling", prefix, f"error: {e}"])

        e = self.create_error(e)

        if isinstance(e, CloseError):
            self.log_debug("Handling a close error")
            ret = self._handle_close_error(e=e, **kwargs)

        else:
            with self.transaction(**kwargs) as connection:
                kwargs["connection"] = connection

                if isinstance(e, UniqueError):
                    self.log_debug("Handling a unique error")
                    ret = self._handle_unique_error(e=e, **kwargs)

                elif isinstance(e, FieldError):
                    self.log_debug("Handling a field error")
                    ret = self._handle_field_error(e=e, **kwargs)

                elif isinstance(e, TableError):
                    self.log_debug("Handling a table error")
                    ret = self._handle_table_error(e=e, **kwargs)

                else:
                    self.log_debug("Handling a general error")
                    ret = self._handle_general_error(e=e, **kwargs)

        if ret:
            self.log_info(["Successfully handled", prefix, "error"])

        else:
            self.log_warning(["Failed to handle", prefix, "error"])

        return ret

    def raise_error(self, e, **kwargs):
        """raises e

        :param e: Exception, if a built-in exception then it's raised, if any other
            error then it will be wrapped in an InterfaceError
        """
        e2 = self.create_error(e, **kwargs)
        if e2 is not e:
            raise e2 from e
        else:
            raise e

    def create_error(self, e, **kwargs):
        """create the error that you want to raise, this gives you an opportunity
        to customize the error

        allow python's built in errors to filter up through
        https://docs.python.org/2/library/exceptions.html

        :param e: Exception, this error will be wrapped in an InterfaceError (or
            whatever is in kwargs["error_class"] if it isn't already an instance
            of that class
        :param **kwargs:
            - error_class: InterfaceError
            - error_module: module, the dbapi module
        """
        error_class = kwargs.get("error_class", InterfaceError)
        if not isinstance(e, error_class):
            if not hasattr(builtins, e.__class__.__name__):
                if "error_module" in kwargs:
                    if kwargs["error_module"].__name__ in e.__class__.__module__:
                        e = error_class(e)

                else:
                    e = error_class(e)

        return e

